from abc import abstractmethod
from typing import Dict, List, Union

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.devices.base import Device


class StorageDevice(Device):
    @abstractmethod
    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        pass


class StorageEl(StorageDevice):
    """Electric storage (battery)"""

    carrier_in = ["el"]
    carrier_out = ["el"]
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceStorageElData,  # Fixme: Correct?
        carrier_data_dict: Dict[str, dto.CarrierElData],  # Fixme: Correct?
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def _rules(self, pyomo_model: pyo.Model, t: int, i: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        dev = self.id
        dev_data: dto.DeviceStorageElData = self.dev_data
        time_delta_minutes = pyomo_model.paramTimestepDeltaMinutes
        time_reserve_minutes = pyomo_model.paramTimeStorageReserveMinutes

        if i == 1:
            # energy balance
            # (el_in*eta - el_out/eta)*dt = delta storage
            # eta = efficiency charging and discharging
            delta_t = time_delta_minutes / 60  # hours
            lhs = (
                pyomo_model.varDeviceFlow[dev, "el", "in", t] * dev_data.eta
                - pyomo_model.varDeviceFlow[dev, "el", "out", t] / dev_data.eta
            ) * delta_t
            if t > 0:
                Eprev = pyomo_model.varDeviceStorageEnergy[dev, t - 1]
            else:
                Eprev = pyomo_model.paramDeviceEnergyInitially[dev]
            rhs = pyomo_model.varDeviceStorageEnergy[dev, t] - Eprev
            return lhs == rhs
        elif i == 2:
            # energy storage limit
            ub = dev_data.E_max
            lb = dev_data.E_min
            # Specifying Emin may be useful e.g. to impose that battery
            # should only be used for reserve
            # BUT - probably better to specify an energy depletion cost
            # instead (Ecost) - to allow using battery (at a cost) if required
            return pyo.inequality(lb, pyomo_model.varDeviceStorageEnergy[dev, t], ub)
        elif i == 3:
            # return pyo.Constraint.Skip # unnecessary -> generic Pmax/min constraints
            # discharging power limit
            ub = dev_data.flow_max
            return pyomo_model.varDeviceFlow[dev, "el", "out", t] <= ub
        elif i == 4:
            # charging power limit - required because the generic flow max/min constraint
            # concerns dis-charging [dev,"el","out",t], cf get_flow_var()
            if dev_data.flow_min is not None:
                ub = -dev_data.flow_min
            else:
                # assume max charging power is the same as discharging power (Pmax)
                ub = dev_data.flow_max  # <- see generic Pmax/min constr
            return pyomo_model.varDeviceFlow[dev, "el", "in", t] <= ub
        elif i == 5:
            # Constraint 5-8: varDeviceStoragePmax = min{Pmax,E/dt}
            # ref: https://or.stackexchange.com/a/1174
            lhs = pyomo_model.varDeviceStoragePmax[dev, t]
            rhs = dev_data.flow_max
            return lhs <= rhs
        elif i == 6:
            lhs = pyomo_model.varDeviceStoragePmax[dev, t]
            # Parameter specifying for how long the power needs to be
            # sustained to count as reserve (e.g. similar to GT startup time)
            dt_hours = time_reserve_minutes / 60
            rhs = pyomo_model.varDeviceStorageEnergy[dev, t] / dt_hours
            return lhs <= rhs
        elif i == 7:
            bigM = 10 * dev_data.flow_max
            lhs = pyomo_model.varDeviceStoragePmax[dev, t]
            rhs = dev_data.flow_max - bigM * (1 - pyomo_model.varStorY[dev, t])
            return lhs >= rhs
        elif i == 8:
            dt_hours = time_reserve_minutes / 60
            bigM = 10 * dev_data.E_max / dt_hours
            lhs = pyomo_model.varDeviceStoragePmax[dev, t]
            rhs = pyomo_model.varDeviceStorageEnergy[dev, t] / dt_hours - bigM * pyomo_model.varStorY[dev, t]
            return lhs >= rhs
        elif i == 9:
            # constraint on storage end vs start
            # Adding this does not seem to improve result (not lower CO2)
            if (dev_data.E_end is not None) and (t == pyomo_model.setHorizon.last()):
                lhs = pyomo_model.varDeviceStorageEnergy[dev, t]
                # rhs = model.varDeviceStorageEnergy[dev,0] # end=start
                rhs = dev_data.E_end
                return lhs == rhs
            else:
                return pyo.Constraint.Skip

    def define_constraints(self, pyomo_model: pyo.Model) -> List[pyo.Constraint]:
        """Specifies the list of constraints for the device"""

        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr = pyo.Constraint(pyomo_model.setHorizon, pyo.RangeSet(1, 9), rule=self._rules)
        # add constraints to model:
        setattr(pyomo_model, "constr_{}_{}".format(self.id, "misc"), constr)
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int) -> float:
        return pyomo_model.varDeviceFlow[self.id, "el", "out", t]

    def get_max_flow(self, pyomo_model: pyo.Model, t: int) -> float:
        # available power may be limited by energy in the storage
        # charging also contributes (can be reversed)
        # (it can go to e.g. -2 MW to +2MW => 4 MW,
        # even if Pmax=2 MW)
        maxValue = pyomo_model.varDeviceStoragePmax[self.id, t] + pyomo_model.varDeviceFlow[self.id, "el", "in", t]
        return maxValue

    def compute_cost_for_depleted_storage(self, pyomo_model: pyo.Model, timesteps: List[int]):
        stor_cost = 0
        dev_data = self.dev_data
        if dev_data.E_cost is not None:
            stor_cost = dev_data.E_cost
        Emax = dev_data.E_max
        storCost = 0
        for t in timesteps:
            varE = pyomo_model.varDeviceStorageEnergy[self.id, t]
            storCost += stor_cost * (Emax - varE)
        return storCost


class StorageHydrogen(StorageDevice):
    "Hydrogen storage"
    carrier_in = ["hydrogen"]
    carrier_out = ["hydrogen"]
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceStorageHydrogenData,  # Fixme: Correct?
        carrier_data_dict: Dict[str, dto.CarrierHydrogenData],  # Fixme: Correct?
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def _rules(self, pyomo_model: pyo.Model, t: int, i: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        dev = self.id
        dev_data: dto.DeviceStorageHydrogenData = self.dev_data
        # param_hydrogen = self.optimiser.all_carriers["hydrogen"]
        time_delta_minutes = pyomo_model.paramTimestepDeltaMinutes
        if i == 1:
            # energy balance (delta E = in - out) (energy in Sm3)
            delta_t = time_delta_minutes * 60  # seconds
            eta = 1
            if dev_data.eta is not None:
                eta = dev_data.eta
            lhs = (
                pyomo_model.varDeviceFlow[dev, "hydrogen", "in", t] * eta
                - pyomo_model.varDeviceFlow[dev, "hydrogen", "out", t] / eta
            ) * delta_t
            if t > 0:
                Eprev = pyomo_model.varDeviceStorageEnergy[dev, t - 1]
            else:
                Eprev = pyomo_model.paramDeviceEnergyInitially[dev]
            rhs = pyomo_model.varDeviceStorageEnergy[dev, t] - Eprev
            return lhs == rhs
        elif i == 2:
            # energy storage limit
            ub = dev_data.E_max
            lb = dev_data.E_min
            return pyo.inequality(lb, pyomo_model.varDeviceStorageEnergy[dev, t], ub)
        elif i == 3:
            # Constraint 3 and 4: to represent absolute value in obj.function
            # see e.g. http://lpsolve.sourceforge.net/5.1/absolute.htm
            #
            # deviation from target and absolute value at the end of horizon
            # TODO: Harald: is there any reason to penalise _positive_ deviation from the target?
            # Xprime>(E_end-E_target)
            # should we instead use
            # Xprime >= 0 (we still need the lower limit (or bound) to avoid negative cost)
            if t != pyomo_model.setHorizon.last():
                return pyo.Constraint.Skip
            Xprime = pyomo_model.varDeviceStorageDeviationFromTarget[dev]
            # profile = model.paramDevice[dev]['target_profile']
            target_value = pyomo_model.paramDeviceEnergyTarget[dev]
            deviation = pyomo_model.varDeviceStorageEnergy[dev, t] - target_value
            return Xprime >= deviation
        elif i == 4:
            # deviation from target and absolute value at the end of horizon
            if t != pyomo_model.setHorizon.last():
                return pyo.Constraint.Skip
            Xprime = pyomo_model.varDeviceStorageDeviationFromTarget[dev]
            # profile = model.paramDevice[dev]['target_profile']
            target_value = pyomo_model.paramDeviceEnergyTarget[dev]
            deviation = pyomo_model.varDeviceStorageEnergy[dev, t] - target_value
            return Xprime >= -deviation
        else:
            raise ValueError(f"Argument i must be 1, 2, 3 or 4. {i} was given.")

    def define_constraints(self, pyomo_model: pyo.Model) -> List[pyo.Constraint]:
        """Specifies the list of constraints for the device"""

        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr = pyo.Constraint(pyomo_model.setHorizon, pyo.RangeSet(1, 4), rule=self._rules)
        # add constraints to model:
        setattr(pyomo_model, "constr_{}_{}".format(self.id, "misc"), constr)
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "hydrogen", "out", t]

    def compute_cost_for_depleted_storage(self, pyomo_model: pyo.Model, timesteps: List[int]):
        return self.compute_costForDepletedStorage_alt2(pyomo_model, timesteps)

    def compute_costForDepletedStorage_alt1(self, pyomo_model: pyo.Model, timesteps: List[int]):
        # cost if storage level at end of optimisation deviates from
        # target profile (user input based on expectations)
        # absolute value of deviation (filling too much also a cost)
        # - avoids over-filling storage
        dev = self.id
        dev_data = self.dev_data
        deviation = pyomo_model.varDeviceStorageDeviationFromTarget[dev]
        stor_cost = dev_data.E_cost * deviation
        return stor_cost

    def compute_costForDepletedStorage_alt2(self, pyomo_model: pyo.Model, timesteps: List[int]):
        # cost rate kr/s
        # Cost associated with deviation from target value
        # below target = cost, above target = benefit   => gives signal to fill storage
        dev_data = self.dev_data
        E_target = dev_data.E_max
        E_target = pyomo_model.paramDeviceEnergyTarget[self.id]
        t_end = timesteps.last()
        varE = pyomo_model.varDeviceStorageEnergy[self.id, t_end]
        stor_cost = dev_data.E_cost * (E_target - varE)
        # from cost (kr) to cost rate (kr/s):
        stor_cost = stor_cost / len(timesteps)
        return stor_cost



from typing import Dict, List, Union

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.devices.base import Device


class WellProduction(Device):
    "Production well (wellstream source)"

    carrier_in = []
    carrier_out = ["wellstream"]
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceWellProductionData,  # Fixme: Correct?
        carrier_data_dict: Dict[str, dto.CarrierWellstreamData],  # Fixme: Correct?
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def _rule_well_production(self, pyomo_model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        dev_data = self.dev_data
        node = dev_data.node_id
        lhs = pyomo_model.varPressure[(node, "wellstream", "out", t)]
        rhs = dev_data.wellhead_pressure
        return lhs == rhs

    def define_constraints(self, pyomo_model: pyo.Model) -> List[pyo.Constraint]:
        """Specifies the list of constraints for the device"""
        # No specific constraints, use only generic ones:
        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr_well = pyo.Constraint(pyomo_model.setHorizon, rule=self._rule_well_production)
        # add constraint to model:
        setattr(
            pyomo_model,
            "constr_{}_{}".format(self.id, "pressure"),
            constr_well,
        )
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int) -> float:
        return pyomo_model.varDeviceFlow[self.id, "wellstream", "out", t]


class WellGasLift(Device):
    "Production well with gas lift"

    carrier_in = ["gas"]
    carrier_out = ["gas", "oil", "water"]
    serial = ["gas"]

    def __init__(
        self,
        dev_data: dto.DeviceWellGasLiftData,  # Fixme: Correct?
        carrier_data_dict: Dict[str, dto.CarrierGasData],  # Fixme: Correct?
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def _rule_gaslift(
        self, pyomo_model: pyo.Model, carrier: dto.CarrierGasData, t: int, i: int
    ) -> Union[pyo.Expression, pyo.Constraint.Skip]:

        # flow from reservoir (equals flow out minus gas injection)
        dev = self.id
        dev_data = self.dev_data
        Q_reservoir = (
            sum(pyomo_model.varDeviceFlow[dev, c, "out", t] for c in ["gas", "oil", "water"])
            - pyomo_model.varDeviceFlow[dev, "gas", "in", t]
        )
        node = dev_data.node_id
        if i == 1:
            # output pressure is fixed
            lhs = pyomo_model.varPressure[(node, carrier, "out", t)]
            rhs = dev_data.separator_pressure
            return lhs == rhs
        elif i == 2:
            # output flow per comonent determined by GOR and WC
            GOR = dev_data.gas_oil_ratio
            WC = dev_data.water_cut
            comp_oil = (1 - WC) / (1 + GOR - GOR * WC)
            comp_water = WC / (1 + GOR - GOR * WC)
            comp_gas = GOR * (1 - WC) / (1 + GOR * (1 - WC))
            comp = {"oil": comp_oil, "gas": comp_gas, "water": comp_water}
            lhs = pyomo_model.varDeviceFlow[dev, carrier, "out", t]
            if carrier == "gas":
                lhs -= pyomo_model.varDeviceFlow[dev, carrier, "in", t]
            rhs = comp[carrier] * Q_reservoir
            return lhs == rhs
        elif i == 3:
            # gas injection rate vs proudction rate (determines input gas)
            # gas injection rate vs OIL proudction rate (determines input gas)
            if carrier == "gas":
                lhs = pyomo_model.varDeviceFlow[dev, "gas", "in", t]
                # rhs = model.paramDevice[dev]['f_inj']*Q_reservoir
                rhs = dev_data.f_inj * pyomo_model.varDeviceFlow[dev, "oil", "out", t]
                return lhs == rhs
            else:
                return pyo.Constraint.Skip
        elif i == 4:
            # gas injection pressure is fixed
            if carrier == "gas":
                lhs = pyomo_model.varPressure[(node, carrier, "in", t)]
                rhs = dev_data.injection_pressure
                return lhs == rhs
            else:
                return pyo.Constraint.Skip
        else:
            raise ValueError(f"Argument i must be 1, 2, 3 or 4. {i} was given.")

    def define_constraints(self, pyomo_model: pyo.Model) -> List[pyo.Constraint]:
        """Specifies the list of constraints for the device"""
        # No specific constraints, use only generic ones:
        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr_gaslift = pyo.Constraint(
            self.carrier_out,
            pyomo_model.setHorizon,
            pyo.RangeSet(1, 4),
            rule=self._rule_gaslift,
        )
        # add constraint to model:
        setattr(
            pyomo_model,
            "constr_{}_{}".format(self.id, "gaslift"),
            constr_gaslift,
        )
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int) -> float:
        dev = self.id
        # flow from reservoir (out minus in)
        flow = (
            pyomo_model.varDeviceFlow[dev, "oil", "out", t]
            + pyomo_model.varDeviceFlow[dev, "gas", "out", t]
            - pyomo_model.varDeviceFlow[dev, "gas", "in", t]
            + pyomo_model.varDeviceFlow[dev, "water", "out", t]
        )
        return flow



import pandas as pd
from pyomo import environ as pyo

try:
    from matplotlib import pyplot as plt
except ImportError:
    raise ImportError("In order to run this plotting module you need to install matplotlib.")


def plot_device_power_last_optimisation_1(mc, device, filename=None):
    model = mc.instance
    devname = model.paramDevice[device]["name"]
    maxP = model.paramDevice[device]["Pmax"]
    plt.figure(figsize=(12, 4))
    ax = plt.gca()
    # plt.title("Results from last optimisation")
    plt.xlabel("Timestep in planning horizon")
    plt.ylabel("Device power (MW)")
    label_profile = ""
    if "profile" in model.paramDevice[device]:
        profile = model.paramDevice[device]["profile"]
        dfa = mc.getProfiles(profile)
        dfa = dfa * maxP
        dfa[profile].plot(ax=ax, label="available")
        label_profile = "({})".format(profile)

    dfIn = pd.DataFrame.from_dict(model.varDeviceFlow.get_values(), orient="index")
    dfIn.index = pd.MultiIndex.from_tuples(dfIn.index, names=("device", "carrier", "terminal", "time"))
    dfIn = dfIn[0].dropna()
    for carr in dfIn.index.levels[1]:
        for term in dfIn.index.levels[2]:
            mask = (dfIn.index.get_level_values(1) == carr) & (dfIn.index.get_level_values(2) == term)
            df_this = dfIn[mask].unstack(0).reset_index()
            if device in df_this:
                df_this[device].plot(
                    ax=ax,
                    linestyle="-",
                    drawstyle="steps-post",
                    marker=".",
                    label="actual ({} {})".format(carr, term),
                )
    ax.legend(loc="upper left")  # , bbox_to_anchor =(1.01,0),frameon=False)

    dfE = pd.DataFrame.from_dict(model.varDeviceStorageEnergy.get_values(), orient="index")
    dfE.index = pd.MultiIndex.from_tuples(dfE.index, names=("device", "time"))
    dfE = dfE[0].dropna()
    df_this = dfE.unstack(0)
    # shift by one because storage at t is storage value _after_ t
    # (just before t+1)
    df_this.index = df_this.index + 1
    if device in df_this:
        ax2 = ax.twinx()
        ax2.set_ylabel("Energy (MWh)", color="red")
        ax2.tick_params(axis="y", labelcolor="red")
        df_this[device].plot(ax=ax2, linestyle="--", color="red", label="storage".format(carr, term))  # noqa
        ax2.legend(
            loc="upper right",
        )
    ax.set_xlim(left=0)

    plt.title("{}:{} {}".format(device, devname, label_profile))
    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")


def plot_device_sum_power_last_optimisation(model: pyo.Model, carrier="el", filename=None):
    """Plot power schedule over planning horizon (last optimisation)"""

    df = pd.DataFrame.from_dict(model.varDeviceFlow.get_values(), orient="index")
    df.index = pd.MultiIndex.from_tuples(df.index, names=("device", "carrier", "inout", "time"))

    # separate out in out
    df = df[0].unstack(level=2)
    df = df.fillna(0)
    df = df["out"] - df["in"]
    df = df.unstack(level=1)
    dfprod = df[df > 0][carrier].dropna()
    dfcons = df[df < 0][carrier].dropna()
    #        dfprod = df[df>=0].unstack(level=1)[carrier]
    #        dfprod = dfprod[dfprod>0]
    #        dfcons = df[df<0].unstack(level=1)[carrier]
    #        dfcons = dfcons[dfcons<0]

    df_info = pd.DataFrame.from_dict(dict(model.paramDevice.items())).T
    labels = df_info.index.astype(str) + "_" + df_info["name"]

    # plt.figure(figsize=(12,4))
    fig, axes = plt.subplots(nrows=2, ncols=1, figsize=(12, 8))
    dfprod.unstack(level=0).rename(columns=labels).plot.area(ax=axes[0], linewidth=0)
    # reverse axes to match stack order
    handles, lgnds = axes[0].get_legend_handles_labels()
    axes[0].legend(
        handles[::-1],
        lgnds[::-1],
        loc="lower left",
        bbox_to_anchor=(1.01, 0),
        frameon=False,
    )
    axes[0].set_ylabel("Produced power (MW)")
    axes[0].set_xlabel("")

    dfcons.unstack(level=0).rename(columns=labels).plot.area(ax=axes[1])
    axes[1].legend(loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)
    axes[1].set_ylabel("Consumed power (MW)")

    axes[1].set_xlabel("Timestep in planning horizon")
    plt.suptitle("Result from last optimisation")
    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")


def plot_rmission_tate_last_optimisation(model: pyo.Model, filename=None):
    devices = model.setDevice
    timesteps = model.setHorizon
    df_info = pd.DataFrame.from_dict(dict(model.paramDevice.items())).T
    labels = df_info.index.astype(str) + "_" + df_info["name"]

    df = pd.DataFrame(index=timesteps, columns=devices)
    for d in devices:
        for t in timesteps:
            co2 = Multicarrier.compute_CO2(model, devices=[d], timesteps=[t])  # noqa: Fixme: This is broken.
            df.loc[t, d] = pyo.value(co2)  # noqa: Fixme: This is broken.
    plt.figure(figsize=(12, 4))
    ax = plt.gca()
    df.loc[:, ~(df == 0).all()].rename(columns=labels).plot.area(ax=ax, linewidth=0)
    plt.xlabel("Timestep")
    plt.ylabel("Emission rate (kgCO2/s)")
    ax.legend(loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)
    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")


# def plotDevicePowerLastOptimisation(model,devices='all',filename=None):
#     """Plot power schedule over planning horizon (last optimisation)"""
#     if devices=='all':
#         devices = list(model.setDevice)
#     varPower = model.getDevicePower()
#     df = pd.DataFrame.from_dict(model.varDevicePower.get_values(),
#                                 orient="index")
#     df.index = pd.MultiIndex.from_tuples(df.index,names=('device','time'))
#     df = df[0].unstack(level=0)
#     df_info = pd.DataFrame.from_dict(dict(model.paramDevice.items())).T
#
#     plt.figure(figsize=(12,4))
#     ax=plt.gca()
#     df[devices].plot(ax=ax)
#     labels = (df_info.loc[devices].index.astype(str)
#               +'_'+df_info.loc[devices,'name'])
#     plt.legend(labels,loc='lower left', bbox_to_anchor =(1.01,0),
#                frameon=False)
#     plt.xlabel("Timestep")
#     plt.ylabel("Device power (MW)")
#     if filename is not None:
#         plt.savefig(filename,bbox_inches = 'tight')
#



import logging
from typing import Optional, Sequence, Tuple

import pandas as pd
import pyomo.environ as pyo
from numpy import float64

from oogeso import dto

from .optimiser import OptimisationModel

logger = logging.getLogger(__name__)

# get progress bar:
try:
    from tqdm import trange

    HAS_TQDM = True
except ImportError:
    logger.info("Consider installing tqdm to get progress bar")
    trange = range
    HAS_TQDM = False


class Simulator:
    """Main class for Oogeso energy system simulations"""

    def __init__(self, data: dto.EnergySystemData):
        """Create Simulator object

        Parameters
        ----------
        data : EnergySystemData
            Data object holding information about the system (nodes, edges, devices, profiles)
            and parameter settings
        """

        # Optimisation model object:
        self.optimiser = OptimisationModel(data=data)

        self.result_object = None

        # Storing time-series profiles:
        _df_profiles_forecast = pd.DataFrame()
        _df_profiles_nowcast = pd.DataFrame()
        for prof in data.profiles:
            _df_profiles_forecast[prof.id] = prof.data
            if prof.data_nowcast is not None:
                _df_profiles_nowcast[prof.id] = prof.data_nowcast
        self.profiles = {
            "nowcast": _df_profiles_nowcast,
            "forecast": _df_profiles_forecast,
        }

    #    def setOptimiser(self, optimiser):
    #        self.optimiser = optimiser

    def run_simulation(
        self,
        solver: str,
        solver_executable: Optional[str] = None,
        solver_options: Optional[dict] = None,
        time_range: Tuple[int, int] = None,
        time_limit: Optional[int] = None,
        return_variables: Optional[Sequence[str]] = None,
        store_duals: Optional[dict] = None,
        write_yaml: bool = False,
    ) -> dto.SimulationResult:
        """Solve problem over many timesteps

        Parameters
        ----------
        solver : string
            Name of solver ("cbc", "gurobi", or others)
        solver_executable : string
            Path to executable
        solver_options : dict
            Solver-specific options passed on to solver (for fine-tuning performance)
        time_range : [int,int]
            Limit to this number of timesteps
        time_limit : int
            Time limit spent on each optimisation (sec)
        return_variables : list
            List of variables to return. Default (None) returns all variables
        store_duals : dict
            Store dual values of constraints. The dictionary contains a
            key:value list where value is a new dictionary specifying the
            constraint and indices. None in the index is replaced by time
            Example:
            store_duals = {'elcost':{constr=constrDevicePmin, indx:('util',None)}
        write_yaml : boolean
            Whether to save problem to yaml file (for debugging)

        Available return variables are:
            "device_flow",
            "device_is_prep",
            "device_is_on",
            "device_starting",
            "device_stopping",
            "device_storage_energy",
            "device_storage_pmax",
            "edge_flow",
            "edge_loss",
            "terminal_flow",
            "terminal_pressure",
            "el_voltage_angle",
            "penalty",
            "el_reserve",
            "el_backup",
            "export_revenue",
            "co2_rate",
            "co2_intensity",
            "co2_rate_per_dev",
            "duals",

        """

        steps = self.optimiser.optimisation_parameters.optimisation_timesteps
        horizon = self.optimiser.optimisation_parameters.planning_horizon
        if time_limit is not None:
            logger.debug("Using solver timelimit=%s", time_limit)
        if time_range is None:
            # use the entire timeseries
            time_start = 0
            time_end = self.profiles["forecast"].index.max() + 1 - horizon
        else:
            time_start = time_range[0]
            time_end = time_range[1]

        result_object = dto.SimulationResult(
            profiles_nowcast=self.profiles["nowcast"],
            profiles_forecast=self.profiles["forecast"],
        )
        self.result_object = result_object

        first = True
        for step in trange(time_start, time_end, steps):
            if not HAS_TQDM:
                # no progress bar
                logger.info("Solving timestep=%s", step)
            # 1. Update problem formulation
            self.optimiser.updateOptimisationModel(step, first=first, profiles=self.profiles)
            # 2. Solve for planning horizon
            self.optimiser.solve(
                solver=solver,
                solver_executable=solver_executable,
                solver_options=solver_options,
                write_yaml=write_yaml,
                time_limit=time_limit,
            )
            # 3. Save results (for later analysis)
            new_results = self._save_optimisation_result(step, return_variables, store_duals)
            result_object.append_results(new_results)
            first = False

        return result_object

    def _save_optimisation_result(
        self,
        timestep: int,
        return_variables: Optional[Sequence[str]] = None,
        store_duals: Optional[dict] = None,
    ) -> dto.SimulationResult:
        """extract results of optimisation for later analysis"""

        pyomo_instance = self.optimiser
        timelimit = self.optimiser.optimisation_parameters.optimisation_timesteps

        return_all = False
        if not return_variables:
            return_all = True
        else:
            logger.debug("Storing only a subset of the data generated.")

        # Retrieve variable values as dictionary with pandas series
        res = self.optimiser.extract_all_variable_values(timelimit, timestep)

        if (return_all or "dfDuals" in return_variables) and (store_duals is not None):
            # Save dual values
            # store_duals = {
            #   'elcost': {'constr':'constrDevicePmin','indx':('util',None)}
            #   }
            horizon_steps = self.optimiser.optimisation_parameters.planning_horizon
            df_duals = pd.DataFrame(columns=store_duals.keys(), index=range(timestep, timestep + timelimit))
            for key, val in store_duals.items():
                # vrs=('util',None)
                vrs = val["indx"]
                constr = getattr(pyomo_instance, val["constr"])
                logger.info(constr)
                # sumduals = 0
                for t in range(timelimit):
                    # Replace None by the timestep, ('util',None) -> ('util',t)
                    vrs1 = tuple(x if x is not None else t for x in vrs)
                    logger.info(vrs1)
                    logger.info(constr[vrs1])
                    dual = pyomo_instance.dual[constr[vrs1]]
                    # The dual gives the improvement in the objective function
                    # if the constraint is relaxed by one unit.
                    # The units of the dual values are the units of the
                    # objective function divided by the units of the constraint.
                    #
                    # A constraint is for a single timestep, whereas the
                    # objective function averages over all timesteps in the
                    # optimisation horizon. To get the improvement of relaxing
                    # the constraint not just in the single timestep, but in
                    # all timesteps we therefore scale up the dual value
                    dual = dual * horizon_steps
                    df_duals.loc[timestep + t, key] = dual
        else:
            df_duals = None

        # CO2 emission rate per device:
        if return_all or "co2_rate_per_dev" in return_variables:
            df_co2_rate_dev = pd.DataFrame(
                index=range(timestep, timestep + timelimit),
                columns=pyomo_instance.setDevice,
            )
            for d in pyomo_instance.setDevice:
                for t in range(timelimit):
                    co2_dev = self.optimiser.compute_CO2(pyomo_instance, devices=[d], timesteps=[t])
                    df_co2_rate_dev.loc[t + timestep, d] = pyo.value(co2_dev)
            # change to multi-index series
            df_co2_rate_dev = df_co2_rate_dev.stack()
            df_co2_rate_dev.index.rename(["time", "device"], inplace=True)
        else:
            df_co2_rate_dev = None

        # CO2 emission rate (sum)
        if return_all or "co2_rate" in return_variables:
            df_co2_rate_sum = pd.Series(dtype=float64, index=range(timestep, timestep + timelimit))
            for t in range(timelimit):
                df_co2_rate_sum.loc[t + timestep] = pyo.value(self.optimiser.compute_CO2(pyomo_instance, timesteps=[t]))
            df_co2_rate_sum.index.rename("time", inplace=True)
        else:
            df_co2_rate_sum = None

        # CO2 emission intensity (sum)
        if return_all or "co2_intensity" in return_variables:
            df_co2intensity = pd.Series(dtype=float64, index=range(timestep, timestep + timelimit))
            for t in range(timelimit):
                df_co2intensity.loc[t + timestep] = pyo.value(
                    self.optimiser.compute_CO2_intensity(pyomo_instance, timesteps=[t])
                )
            df_co2intensity.index.rename("time", inplace=True)
        else:
            df_co2intensity = None

        # Penalty values per device
        # df_penalty=res["varDevicePenalty"], # this does not include start/stop penalty
        if return_all or "penalty" in return_variables:
            df_penalty = pd.DataFrame(
                dtype=float64,
                index=range(timestep, timestep + timelimit),
                columns=pyomo_instance.setDevice,
            )
            for d, dev in self.optimiser.all_devices.items():
                for t in range(timelimit):
                    this_penalty = dev.compute_penalty(pyomo_instance, [t])
                    df_penalty.loc[t + timestep, d] = pyo.value(this_penalty)
            # change to multi-index series:
            df_penalty = df_penalty.stack()
            df_penalty.index.rename(["time", "device"], inplace=True)
        else:
            df_penalty = None

        # Revenue from exported energy (per carrier)
        if return_all or "export_revenue" in return_variables:
            df_exportRevenue = pd.DataFrame(
                dtype=float64,
                index=range(timestep, timestep + timelimit),
                columns=pyomo_instance.setCarrier,
            )
            for c in pyomo_instance.setCarrier:
                for t in range(timelimit):
                    exportRevenue_dev = self.optimiser.compute_exportRevenue(
                        pyomo_instance, carriers=[c], timesteps=[t]
                    )
                    df_exportRevenue.loc[t + timestep, c] = pyo.value(exportRevenue_dev)
            # change to multi-index series:
            df_exportRevenue = df_exportRevenue.stack()
            df_exportRevenue.index.rename(["time", "carrier"], inplace=True)
        else:
            df_exportRevenue = None

        # Reserve capacity
        if return_all or "el_reserve" in return_variables:
            df_reserve = pd.Series(dtype=float64, index=range(timestep, timestep + timelimit))
            for t in range(timelimit):
                rescap = pyo.value(
                    self.optimiser.all_networks["el"].compute_el_reserve(pyomo_instance, t, self.optimiser.all_devices)
                )
                df_reserve.loc[t + timestep] = rescap
            df_reserve.index.rename("time", inplace=True)
        else:
            df_reserve = None

        # Backup capacity
        if return_all or "el_backup" in return_variables:
            devs_elout = []
            for dev_obj in self.optimiser.all_devices.values():
                if "el" in dev_obj.carrier_out:
                    devs_elout.append(dev_obj.id)
            df_backup = pd.DataFrame(
                dtype=float64,
                index=range(timestep, timestep + timelimit),
                columns=devs_elout,
            )
            for t in range(timelimit):
                for d in devs_elout:
                    rescap = pyo.value(
                        self.optimiser.all_networks["el"].compute_el_reserve(
                            pyomo_instance,
                            t,
                            self.optimiser.all_devices,
                            exclude_device=d,
                        )
                    )
                    df_backup.loc[t + timestep, d] = rescap
            df_backup = df_backup.stack()
            df_backup.index.rename(["time", "device"], inplace=True)
        else:
            df_backup = None

        dfDeviceFlow = res["varDeviceFlow"] if (return_all or "device_flow" in return_variables) else None
        dfDeviceIsOn = res["varDeviceIsOn"] if (return_all or "device_is_on" in return_variables) else None
        dfDeviceIsPrep = res["varDeviceIsPrep"] if (return_all or "device_is_prep" in return_variables) else None
        dfDeviceStarting = res["varDeviceStarting"] if (return_all or "device_starting" in return_variables) else None
        dfDeviceStopping = res["varDeviceStopping"] if (return_all or "device_stopping" in return_variables) else None
        dfDeviceStorageEnergy = (
            res["varDeviceStorageEnergy"] if (return_all or "device_storage_energy" in return_variables) else None
        )
        dfDeviceStoragePmax = (
            res["varDeviceStoragePmax"] if (return_all or "device_storage_pmax" in return_variables) else None
        )
        dfEdgeFlow = res["varEdgeFlow"] if (return_all or "edge_flow" in return_variables) else None
        dfEdgeLoss = res["varEdgeLoss"] if (return_all or "edge_loss" in return_variables) else None
        dfTerminalFlow = res["varTerminalFlow"] if (return_all or "terminal_flow" in return_variables) else None
        dfTerminalPressure = res["varPressure"] if (return_all or "terminal_pressure" in return_variables) else None
        dfElVoltageAngle = res["varElVoltageAngle"] if (return_all or "el_voltage_angle" in return_variables) else None

        result_object = dto.SimulationResult(
            device_flow=dfDeviceFlow,
            device_is_prep=dfDeviceIsPrep,
            device_is_on=dfDeviceIsOn,
            device_starting=dfDeviceStarting,
            device_stopping=dfDeviceStopping,
            device_storage_energy=dfDeviceStorageEnergy,
            device_storage_pmax=dfDeviceStoragePmax,
            edge_flow=dfEdgeFlow,
            edge_loss=dfEdgeLoss,
            terminal_flow=dfTerminalFlow,
            terminal_pressure=dfTerminalPressure,
            el_voltage_angle=dfElVoltageAngle,
            penalty=df_penalty,
            el_reserve=df_reserve,
            el_backup=df_backup,
            export_revenue=df_exportRevenue,
            co2_rate=df_co2_rate_sum,
            co2_intensity=df_co2intensity,
            co2_rate_per_dev=df_co2_rate_dev,
            duals=df_duals,
            profiles_forecast=None,
            profiles_nowcast=None,
        )
        return result_object



from typing import Dict, List, Union

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.devices.base import Device


class GasHeater(Device):
    "Gas heater"
    carrier_in = ["gas"]
    carrier_out = ["heat"]
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceGasHeaterData,
        carrier_data_dict: Dict[str, dto.CarrierGasData],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def _rules(self, pyomo_model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        dev = self.id
        param_dev = self.params  # Fixme: Attribute params is missing
        param_gas = pyomo_model.all_carriers["gas"].params
        # heat out = gas input * energy content * efficiency
        gas_energy_content = param_gas["energy_value"]  # MJ/Sm3
        lhs = pyomo_model.varDeviceFlow[dev, "heat", "out", t]
        rhs = pyomo_model.varDeviceFlow[dev, "gas", "in", t] * gas_energy_content * param_dev["eta"]
        return lhs == rhs

    def define_constraints(self, pyomo_model: pyo.Model):
        """Specifies the list of constraints for the device"""

        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr = pyo.Constraint(pyomo_model.setHorizon, rule=self._rules)
        # add constraint to model:
        setattr(pyomo_model, "constr_{}_{}".format(self.id, "misc"), constr)
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        # using heat output as dimensioning variable
        # (alternative could be to use gas input)
        return pyomo_model.varDeviceFlow[self.id, "heat", "out", t]

    # overriding default
    def compute_CO2(self, pyomo_model: pyo.Model, timesteps: List[int]) -> float:
        """
        Fixme: The variable d and model_pyomo was not set. Changed to model and self.dev_data. Correct?
        """
        param_gas = self.carrier_data["gas"]
        gas_flow_co2 = param_gas.co2_content  # kg/m3
        return sum(pyomo_model.varDeviceFlow[self.dev_data, "gas", "in", t] for t in timesteps) * gas_flow_co2



import logging
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Union

import pyomo.environ as pyo
from pyomo.core import Constraint

from oogeso.core.networks.network_node import NetworkNode
from oogeso.dto import CarrierData, DeviceData, TimeSeriesData

logger = logging.getLogger(__name__)


class Device(ABC):
    """
    Parent class from which all device types derive
    """

    # Common class parameters:
    carrier_in = list()
    carrier_out = list()
    serial = list()

    def __init__(self, dev_data: DeviceData, carrier_data_dict: Dict[str, CarrierData]):
        """Device object constructor"""
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict
        self.node: Optional[NetworkNode] = None
        self._flow_upper_bound: Optional[pyo.Constraint] = None

    def add_node(self, node: NetworkNode) -> None:
        """associate node with device"""
        self.node = node

    def set_init_values(self, pyomo_model: pyo.Model) -> None:
        """Method invoked to specify optimisation problem initial value parameters"""
        # TODO: move this to each subclass instead?
        dev_id = self.id
        dev_data = self.dev_data
        if hasattr(dev_data, "E_init"):
            pyomo_model.paramDeviceEnergyInitially[dev_id] = dev_data.E_init
        if dev_data.start_stop is not None:
            pyomo_model.paramDeviceIsOnInitially[dev_id] = dev_data.start_stop.is_on_init
        if hasattr(dev_data, "P_init"):
            pyomo_model.paramDevicePowerInitially[dev_id] = dev_data.P_init

    def _rule_device_flow_max(self, pyomo_model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        power = self.get_flow_var(pyomo_model, t)
        if power is None:
            return pyo.Constraint.Skip
        max_value = self.get_max_flow(pyomo_model, t)
        expr = power <= max_value
        return expr

    def _rule_device_flow_min(self, pyomo_model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        power = self.get_flow_var(pyomo_model, t)
        if power is None:
            return pyo.Constraint.Skip
        min_value = self.dev_data.flow_min
        if self.dev_data.profile is not None:
            # use an availability profile if provided
            ext_profile = self.dev_data.profile
            min_value = min_value * pyomo_model.paramProfiles[ext_profile, t]
        ison = 1
        if self.dev_data.start_stop is not None:
            ison = pyomo_model.varDeviceIsOn[self.id, t]
        expr = power >= ison * min_value
        return expr

    def _rule_ramp_rate(self, pyomo_model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """power ramp rate limit"""
        dev = self.id
        dev_data = self.dev_data

        # If no ramp limits have been specified, skip constraint
        if self.dev_data.max_ramp_up is None:
            return pyo.Constraint.Skip
        if t > 0:
            p_prev = self.get_flow_var(pyomo_model, t - 1)
        else:
            p_prev = pyomo_model.paramDevicePowerInitially[dev]
        p_this = self.get_flow_var(pyomo_model, t)
        delta_P = p_this - p_prev
        delta_t = pyomo_model.paramTimestepDeltaMinutes
        max_P = dev_data.flow_max
        max_neg = -dev_data.max_ramp_down * max_P * delta_t
        max_pos = dev_data.max_ramp_up * max_P * delta_t
        expr = pyo.inequality(max_neg, delta_P, max_pos)
        return expr

    def _rule_startup_shutdown(self, model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """startup/shutdown constraint
        connecting starting, stopping, preparation, online stages of GTs"""
        dev = self.id
        T_delay_min = self.dev_data.start_stop.delay_start_minutes
        time_delta_minutes = model.paramTimestepDeltaMinutes
        # Delay in time-steps, rounding down.
        # example: time_delta = 5 min, delay_start_minutes= 8 min => T_delay=1
        T_delay = int(T_delay_min / time_delta_minutes)
        prev_part = 0
        if t >= T_delay:
            # prev_part = sum( model.varDeviceStarting[dev,t-tau]
            #    for tau in range(0,T_delay) )
            prev_part = model.varDeviceStarting[dev, t - T_delay]
        else:
            # NOTE: for this to work as intended, may need to reconstruct constraint
            # pyo.value(...) needed in pyomo v6
            prep_init = pyo.value(model.paramDevicePrepTimestepsInitially[dev])
            if prep_init + t == T_delay:
                prev_part = 1
        if t > 0:
            is_on_prev = model.varDeviceIsOn[dev, t - 1]
        else:
            is_on_prev = model.paramDeviceIsOnInitially[dev]
        lhs = model.varDeviceIsOn[dev, t] - is_on_prev
        rhs = prev_part - model.varDeviceStopping[dev, t]
        return lhs == rhs

    def _rule_startup_delay(self, pyomo_model: pyo.Model, t: int) -> Union[bool, pyo.Constraint, pyo.Constraint.Skip]:
        """startup delay/preparation for GTs"""
        dev = self.id
        time_delta_minutes = pyomo_model.paramTimestepDeltaMinutes
        T_delay_min = self.dev_data.start_stop.delay_start_minutes
        # Delay in time-steps, rounding down.
        # example: time_delta = 5 min, startupDelay= 8 min => T_delay=1
        T_delay = int(T_delay_min / time_delta_minutes)
        if T_delay == 0:
            return pyo.Constraint.Skip
        # determine if was in preparation previously
        # dependent on value - so must reconstruct constraint each time
        steps_prev_prep = pyo.value(pyomo_model.paramDevicePrepTimestepsInitially[dev])
        if steps_prev_prep > 0:
            prev_is_prep = 1
        else:
            prev_is_prep = 0

        prev_part = 0
        if t < T_delay - steps_prev_prep:
            prev_part = prev_is_prep
        tau_range = range(0, min(t + 1, T_delay))
        lhs = pyomo_model.varDeviceIsPrep[dev, t]
        rhs = sum(pyomo_model.varDeviceStarting[dev, t - tau] for tau in tau_range) + prev_part
        return lhs == rhs

    def define_constraints(self, pyomo_model: pyo.Model) -> List[Constraint]:
        """Build constraints for the device and add to pyomo model.
        Returns list of constraints that need to be reconstructed between each
        optimisation

        Fixme: Make setattr robust. Possible to set the wrong attribute without knowing.
        """

        list_to_reconstruct = []  # Default

        if self.dev_data.flow_max is not None:
            constrDevicePmax = pyo.Constraint(pyomo_model.setHorizon, rule=self._rule_device_flow_max)
            setattr(
                pyomo_model,
                f"constr_{self.id}_flowMax",
                constrDevicePmax,
            )
        if self.dev_data.flow_min is not None:
            constrDevicePmin = pyo.Constraint(pyomo_model.setHorizon, rule=self._rule_device_flow_min)
            setattr(
                pyomo_model,
                f"constr_{self.id}_flowMin",
                constrDevicePmin,
            )
        if (self.dev_data.max_ramp_up is not None) or (self.dev_data.max_ramp_down is not None):
            constrDevice_ramprate = pyo.Constraint(pyomo_model.setHorizon, rule=self._rule_ramp_rate)
            setattr(
                pyomo_model,
                f"constr_{self.id}_ramprate",
                constrDevice_ramprate,
            )
        if self.dev_data.start_stop is not None:
            constrDevice_startup_shutdown = pyo.Constraint(pyomo_model.setHorizon, rule=self._rule_startup_shutdown)
            setattr(
                pyomo_model,
                f"constr_{self.id}_startstop",
                constrDevice_startup_shutdown,
            )
            constrDevice_startup_delay = pyo.Constraint(pyomo_model.setHorizon, rule=self._rule_startup_delay)
            setattr(
                pyomo_model,
                f"constr_{self.id}_startdelay",
                constrDevice_startup_delay,
            )

            # TODO: Add constraints for minimum up and down-time

            # return list of constraints that need to be reconstructed:
            list_to_reconstruct = [
                constrDevice_startup_shutdown,
                constrDevice_startup_delay,
            ]
        return list_to_reconstruct

    @abstractmethod
    def get_flow_var(self, pyomo_model: pyo.Model, t: int) -> float:
        pass

    def get_max_flow(self, pyomo_model: pyo.Model, t: int) -> float:
        """
        Return available capacity at given time-step.
        This is given by the "flow_max" input parameter, profile value (if any), and
        whether device is on/off.
        """
        max_value = self.dev_data.flow_max
        if self.dev_data.profile is not None:
            ext_profile = self.dev_data.profile
            max_value = max_value * pyomo_model.paramProfiles[ext_profile, t]
        if self.dev_data.start_stop is not None:
            is_on = pyomo_model.varDeviceIsOn[self.id, t]
            max_value = is_on * max_value
        return max_value

    def set_flow_upper_bound(self, profiles: List[TimeSeriesData]) -> None:
        """
        Maximum flow value through entire profile.

        Given as the product of the "flow_max" parameter and profile values.
        """
        ub = self.dev_data.flow_max
        if self.dev_data.profile is not None:
            ext_profile = self.dev_data.profile
            prof_max = None
            for prof in profiles:
                if prof.id == ext_profile:
                    prof_max = max(prof.data)
                    if prof.data_nowcast is not None:
                        prof_nowcast_max = max(prof.data_nowcast)
                        prof_max = max(prof_max, prof_nowcast_max)
                    ub = ub * prof_max
                    break
            if prof_max is None:
                logger.warning(
                    "Profile (%s) defined for device %s was not found",
                    ext_profile,
                    self.dev_data.id,
                )
        self._flow_upper_bound = ub

    def get_flow_upper_bound(self) -> Optional[pyo.Constraint]:
        """Returns the maximum possible flow given capacity and profile"""
        # Used by piecewise linear constraints
        return self._flow_upper_bound

    def compute_export(
        self, pyomo_model: pyo.Model, value: str, carriers: List[CarrierData], timesteps: List[int]
    ) -> float:
        """Compute average export (volume or revenue)

        Parameters:
        -----------
        value : str
            "revenue" (€/s) or "volume" (Sm3oe/s)
        carriers : list of carriers ("gas","oil","el")
        timesteps : list of timesteps

        Fixme: price needs to be handled in subclasses, or price needs to be an Optional attribute of DeviceData.
        """
        carriers_in = self.carrier_in
        carriers_incl = [v for v in carriers if v in carriers_in]
        sumValue = 0
        if not hasattr(self.dev_data, "price"):
            return 0
        for carrier in carriers_incl:
            # flow in m3/s, price in $/m3
            if self.dev_data.price is not None:
                if carrier in self.dev_data.price:
                    inflow = sum(pyomo_model.varDeviceFlow[self.id, carrier, "in", t] for t in timesteps)
                    if value == "revenue":
                        sumValue += inflow * self.dev_data.price[carrier]
                    elif value == "volume":
                        volumefactor = 1
                        if carrier == "gas":
                            volumefactor = 1 / 1000  # Sm3 to Sm3oe
                        sumValue += inflow * volumefactor
        return sumValue

    def compute_el_reserve(self, pyomo_model: pyo.Model, t: int) -> Dict[str, float]:
        """Compute available reserve power from this device

        device parameter "reserve_factor" specifies how large part of the
        available capacity should count towards the reserve (1=all, 0=none)
        """
        rf = 1
        load_reduction = 0
        cap_avail = 0
        p_generating = 0
        if "el" in self.carrier_out:
            # Generators and storage
            max_value = self.get_max_flow(pyomo_model=pyomo_model, t=t)
            if self.dev_data.reserve_factor is not None:
                # safety margin - only count a part of the forecast power
                # towards the reserve, relevant for wind power
                # (equivalently, this may be seen as increaseing the
                # reserve margin requirement)
                reserve_factor = self.dev_data.reserve_factor
                max_value = max_value * reserve_factor
                if reserve_factor == 0:
                    # no reserve contribution
                    rf = 0
            cap_avail = rf * max_value
            p_generating = rf * pyomo_model.varDeviceFlow[self.id, "el", "out", t]
        elif "el" in self.carrier_in:
            # Loads (only consider if resere factor has been set)
            if self.dev_data.reserve_factor is not None:
                # load reduction possible
                f_lr = self.dev_data.reserve_factor
                load_reduction = f_lr * pyomo_model.varDeviceFlow[self.id, "el", "in", t]
        reserve = {
            "capacity_available": cap_avail,
            "capacity_used": p_generating,
            "loadreduction_available": load_reduction,
        }
        return reserve

    def compute_startup_penalty(self, pyomo_model: pyo.Model, timesteps: List[int]) -> float:
        """start/stop penalty - generalised from gas turbine startup cost."""
        penalty = 0
        if self.dev_data.start_stop is not None:
            penalty = (
                sum(pyomo_model.varDeviceStarting[self.id, t] for t in timesteps)
                * self.dev_data.start_stop.penalty_start
                + sum(pyomo_model.varDeviceStopping[self.id, t] for t in timesteps)
                * self.dev_data.start_stop.penalty_stop
            )
        return penalty

    def compute_operating_costs(self, pyomo_model: pyo.Model, timesteps: List[int]) -> float:
        """average operating cost within selected timespan"""
        sum_cost = 0
        if self.dev_data.op_cost is not None:
            op_cost = self.dev_data.op_cost
            for t in pyomo_model.setHorizon:
                var_P = self.get_flow_var(pyomo_model=pyomo_model, t=t)
                sum_cost += op_cost * var_P
        # average per sec (simulation timestep drops out)
        avgCost = sum_cost / len(timesteps)
        return avgCost

    def compute_cost_for_depleted_storage(self, pyomo_model: pyo.Model, timesteps: List[int]):
        return 0

    def compute_penalty(self, pyomo_model: pyo.Model, timesteps: List[int]):
        """Compute average penalty rate (cost, emission or similar per second)
        as defined by penalty functions and start/stop penalties"""
        penalty_rate = 0

        # Fixme: Add Optional penalty function in DeviceData.
        if hasattr(self.dev_data, "penalty_function") and self.dev_data.penalty_function is not None:
            if not hasattr(self, "_penaltyConstraint"):
                logger.warning("Penalty function constraint not impelemented for %s", self.id)
            # Since the penalty function may be nonzero at Pel=0 we need to split up so computed
            # penalty for Pel > 0 only when device is actually online (penalty should be zero when
            # device is offline)
            penalty_offset = 0
            if self.dev_data.start_stop is not None:
                # penalty_offset = penalty(Pel=0)
                penalty_offset = self.dev_data.penalty_function[1][0]
            this_penalty = sum(
                pyomo_model.varDevicePenalty[self.id, "el", "out", t]
                + (pyomo_model.varDeviceIsOn[self.id, t] - 1) * penalty_offset
                for t in timesteps
            )
            # divide by number of timesteps to get average penalty rate (penalty per sec):
            penalty_rate = this_penalty / len(timesteps)

        start_stop_penalty_rate = 0
        if self.dev_data.start_stop is not None:
            # this sums up penalty over all timesteps in horizon:
            start_stop_penalty = self.compute_startup_penalty(pyomo_model, timesteps)
            # get average per second:
            time_delta_minutes = pyomo_model.paramTimestepDeltaMinutes
            timestep_duration_sec = time_delta_minutes * 60
            time_interval_sec = len(timesteps) * timestep_duration_sec
            start_stop_penalty_rate = start_stop_penalty / time_interval_sec

        sum_penalty_rate = penalty_rate + start_stop_penalty_rate
        return sum_penalty_rate

    def compute_CO2(self, pyomo_model: pyo.Model, timesteps: List[int]) -> float:
        """Standard for all devices unless overridden."""
        return 0



from dataclasses import dataclass, fields
from typing import Optional

import pandas as pd


@dataclass
class SimulationResult:
    """Results from oogeso simulation

    The results are stored in a set of multi-index Series, with
    index names indicating what they are:

    device - device identifier
    node - node identifier
    edge - edge identifier
    carrier - network type ("el", "gas", "oil", "water", "hydrogen", "heat")
    terminal - input/output ("in" or "out"),
    time (integer timestep)
    """

    # Input/output flow per device and network type:
    device_flow: Optional[pd.Series] = None
    # Device startup preparation status (boolean):
    device_is_prep: Optional[pd.Series] = None
    # Device on/off status (boolean):
    device_is_on: Optional[pd.Series] = None
    # Device starting status (boolean):
    device_starting: Optional[pd.Series] = None
    # Device stopping status (boolean):
    device_stopping: Optional[pd.Series] = None
    # Energy storage filling level (Sm3 or MJ)
    device_storage_energy: Optional[pd.Series] = None
    # Max available "flow" (power/fluid) from storage (Sm3/s or MW):
    device_storage_pmax: Optional[pd.Series] = None
    # Device assosiated penalty rate (PENALTY_UNIT/s):
    penalty: Optional[pd.Series] = None
    # Flow rate (Sm3/s or MW):
    edge_flow: Optional[pd.Series] = None
    # Loss rate (MW) - only relevant for energy flow (el and heat):
    edge_loss: Optional[pd.Series] = None
    # Voltage angle at node - only relevant for electricity floc computed via dc-pf:
    el_voltage_angle: Optional[pd.Series] = None
    # Pressure at node (MPa):
    terminal_pressure: Optional[pd.Series] = None
    # Direct flow between in and out terminal of node - relevant if there is no device inbetween:
    terminal_flow: Optional[pd.Series] = None
    # Emission rate (sum of all devices) (kgCO2/s):
    co2_rate: Optional[pd.Series] = None
    # Emission rate per device (kgCO2/s):
    co2_rate_per_dev: Optional[pd.Series] = None
    # Revenue rate for exported oil/gas (CURRENCY/s):
    export_revenue: Optional[pd.Series] = None
    # CO2 intensity of exported oil/gas (kgCO2/Sm3oe):
    co2_intensity: Optional[pd.Series] = None
    # Available online electrical reserve capacity (MW):
    el_reserve: Optional[pd.Series] = None
    # Available online electrical backup per device (MW):
    el_backup: Optional[pd.Series] = None
    # Value of duals (associated with constraints)
    duals: Optional[pd.Series] = None
    # Time-series profiles used in simulation (copied from the input)
    profiles_forecast: Optional[pd.DataFrame] = None
    profiles_nowcast: Optional[pd.DataFrame] = None

    def append_results(self, sim_res):
        exclude_list = ["df_profiles_forecast", "df_profiles_forecast"]
        for my_field in fields(self):
            field_name = my_field.name
            if field_name not in exclude_list:
                my_df = getattr(self, field_name)
                other_df = getattr(sim_res, field_name)
                if other_df is not None:
                    setattr(self, field_name, pd.concat([my_df, other_df]).sort_index())



from dataclasses import dataclass, field
from typing import List, Optional, Tuple

from oogeso import dto

# Device types defined as part of "basic":
# DevicePowerSourceData
# DeviceSink_elData
# DeviceStorage_elData


@dataclass
class DeviceSourceElData(dto.DeviceData):
    co2em: Optional[float] = None
    op_cost: Optional[float] = None
    reserve_factor: float = 1  # not used capacity contributes fully to spinning reserve


@dataclass
class DeviceSourceGasData(dto.DeviceData):
    naturalpressure: float = None


@dataclass
class DeviceSourceWaterData(dto.DeviceData):
    naturalpressure: float = None


@dataclass
class DeviceSinkElData(dto.DeviceData):
    pass


@dataclass
class DeviceSinkHeatData(dto.DeviceData):
    pass


@dataclass
class DeviceSinkGasData(dto.DeviceData):
    price: field(default_factory=lambda: {}) = None


@dataclass
class DeviceSinkOilData(dto.DeviceData):
    price: field(default_factory=lambda: {}) = None


@dataclass
class DeviceSinkWaterData(dto.DeviceData):
    price: field(default_factory=lambda: {}) = None
    flow_avg: Optional[float] = None  # required average flow
    max_accumulated_deviation: Optional[float] = None  # buffer size (max accumulated deviation from average)


@dataclass
class DeviceCompressorElData(dto.DeviceData):
    eta: float = None  # efficiency
    Q0: float = None  # nominal flow rate used in linearisation
    temp_in: float = None  # inlet temperature


@dataclass
class DeviceCompressorGasData(dto.DeviceData):
    eta: float = None  # efficiency
    Q0: float = None  # nominal flow rate used in linearisation
    temp_in: float = None  # inlet temperature


@dataclass
class DeviceElectrolyserData(dto.DeviceData):
    eta: float = None  # efficiency
    eta_heat: float = None  # heat recovery efficiency


@dataclass
class DeviceFuelCellData(dto.DeviceData):
    eta: float = None  # efficiency
    eta_heat: float = None  # heat recovery efficiency


@dataclass
class DeviceGasHeaterData(dto.DeviceData):
    pass


@dataclass
class DeviceGasTurbineData(dto.DeviceData):
    fuel_A: float = None
    fuel_B: float = None
    eta_heat: float = None
    #    is_on_init: bool = False
    #    startup_cost: float = None
    #    startup_delay: float = None  # Minutes from activation to power delivery
    #    shutdown_cost: float = None
    reserve_factor: float = 1  # not used capacity contributes fully to spinning reserve


@dataclass
class DeviceHeatPumpData(dto.DeviceData):
    eta: float = None


@dataclass
class DevicePumpOilData(dto.DeviceData):
    eta: float = None  # efficiency


@dataclass
class DevicePumpWaterData(dto.DeviceData):
    eta: float = None


@dataclass
class DeviceSeparatorData(dto.DeviceData):
    el_demand_factor: float = None  # electricity demand factor
    heat_demand_factor: float = None  # heat demand factor


@dataclass
class DeviceSeparator2Data(dto.DeviceData):
    el_demand_factor: float = None  # electricity demand factor
    heat_demand_factor: float = None  # heat demand factor


@dataclass
class DeviceStorageHydrogenData(dto.DeviceData):
    E_max: float = 0  # MWh storage capacity (maximum stored energy)
    E_min: float = 0
    eta: float = 1  # efficiency
    target_profile: Optional[str] = None  # target profile for use of (seasonal) storage
    E_cost: float = 0  # cost for depleting storage
    E_init: float = 0


@dataclass
class DeviceWellProductionData(dto.DeviceData):
    wellhead_pressure: float = None  # 2 # MPa


@dataclass
class DeviceWellGasLiftData(dto.DeviceData):
    gas_oil_ratio: float = None  # 500
    water_cut: float = None  # 0.6
    f_inj: float = None  # 220 # gas injection rate as fraction of production rate
    injection_pressure: float = None  # 20 # MPa
    separator_pressure: float = None  # 2 # MPa


@dataclass
class EdgeHeatData(dto.EdgeData):
    # Heat loss in MW as function of energy transfer in MW:
    power_loss_function: Optional[Tuple[List[float], List[float]]] = None


@dataclass
class EdgeHydrogenData(dto.EdgeData):
    bidirectional: bool = False


@dataclass
class EdgeFluidData(dto.EdgeData):
    # wellstream, oil, water, gas
    pressure_from: float = None
    pressure_to: float = None
    diameter_mm: float = None
    temperature_K: float = None
    height_m: float = 0
    num_pipes: Optional[int] = None
    bidirectional: bool = False
    # allovable relative deviation of pressure from nominal values,
    pressure_from_maxdeviation: Optional[float] = None
    pressure_to_maxdeviation: Optional[float] = None


@dataclass
class EdgeGasData(EdgeFluidData):
    pass


@dataclass
class EdgeOilData(EdgeFluidData):
    pass


@dataclass
class EdgeWellstreamData(EdgeFluidData):
    pass


@dataclass
class EdgeWaterData(EdgeFluidData):
    pass


@dataclass
class CarrierHeatData(dto.CarrierData):
    pass


@dataclass
class CarrierHydrogenData(dto.CarrierData):
    energy_value: float = 13  # MJ/Sm3 (calorific value) -> 13 MJ/Sm3


@dataclass
class CarrierGasData(dto.CarrierData):
    co2_content: float  # kg/Sm3 - see SSB 2016 report -> 2.34 kg/Sm3
    G_gravity: float  # 0.6
    Pb_basepressure_MPa: float  # MPa -> 0.101 # MPa
    R_individual_gas_constant: float  # J/(kg K) -> 500 J/kgK
    Tb_basetemp_K: float  # K -> 288 K = 15 degC
    Z_compressibility: float  # 0.9
    energy_value: float  # MJ/Sm3 (calorific value) -> 40 MJ/Sm3
    k_heat_capacity_ratio: float  # 1.27
    rho_density: float  # kg/m3 -> 0.84 kg/m3
    pressure_method: Optional[str] = "weymouth"  # pressure drop calculation


@dataclass
class CarrierWellstreamData(dto.CarrierData):
    darcy_friction: float = None  # 0.02
    rho_density: float = None  # kg/m3 -> 900 kg/m3
    viscosity: float = None  # kg/(m s) -> 0.0026 kg/(m s)
    pressure_method: Optional[str] = None
    water_cut: float = None
    gas_oil_ratio: float = None


@dataclass
class CarrierOilData(dto.CarrierData):
    darcy_friction: float = None  # 0.02
    rho_density: float = None  # kg/m3 -> 900 kg/m3
    viscosity: float = None  # kg/(m s) -> 0.0026 kg/(m s)
    pressure_method: Optional[str] = "darcy-weissbach"  # pressure drop calculation


@dataclass
class CarrierWaterData(dto.CarrierData):
    darcy_friction: float = None  # 0.02
    rho_density: float = None  # kg/m3 -> 900 kg/m3
    viscosity: float = None  # kg/(m s) -> 0.0026 kg/(m s)
    pressure_method: Optional[str] = "darcy-weissbach"  # pressure drop calculation



from typing import Dict, List, Union

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.devices.base import Device


class GasTurbine(Device):
    "Gas turbine generator"

    carrier_in = ["gas"]
    carrier_out = ["el", "heat"]
    serial = list()

    def __init__(
        self,
        dev_data: dto.DeviceGasTurbineData,
        carrier_data_dict: Dict[str, dto.CarrierGasData],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def _rules_misc(self, model: pyo.Model, t: int, i: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        dev = self.id
        param_gas = self.carrier_data["gas"]
        # el_power = model.varDeviceFlow[dev, "el", "out", t]
        gas_energy_content = param_gas.energy_value  # MJ/Sm3
        if i == 1:
            """turbine el power out vs gas fuel in"""
            # fuel consumption (gas in) is a linear function of el power output
            # fuel = B + A*power
            # => efficiency = power/(A+B*power)
            A = self.dev_data.fuel_A
            B = self.dev_data.fuel_B
            Pmax = self.dev_data.flow_max
            lhs = model.varDeviceFlow[dev, "gas", "in", t] * gas_energy_content / Pmax
            rhs = (
                B * (model.varDeviceIsOn[dev, t] + model.varDeviceIsPrep[dev, t])
                + A * model.varDeviceFlow[dev, "el", "out", t] / Pmax
            )
            return lhs == rhs
        elif i == 2:
            """heat output = (gas energy in - el power out)* heat efficiency"""
            lhs = model.varDeviceFlow[dev, "heat", "out", t]
            rhs = (
                model.varDeviceFlow[dev, "gas", "in", t] * gas_energy_content - model.varDeviceFlow[dev, "el", "out", t]
            ) * self.dev_data.eta_heat
            return lhs == rhs

    def define_constraints(self, pyomo_model: pyo.Model):
        """Specifies the list of constraints for the device"""
        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr = pyo.Constraint(pyomo_model.setHorizon, pyo.RangeSet(1, 2), rule=self._rules_misc)
        setattr(pyomo_model, "constr_{}_{}".format(self.id, "misc"), constr)
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "el", "out", t]

    # overriding default
    def compute_CO2(self, pyomo_model: pyo.Model, timesteps: List[int]) -> float:
        param_gas = self.carrier_data["gas"]
        gasflow_co2 = param_gas.co2_content  # kg/m3

        return sum(pyomo_model.varDeviceFlow[self.id, "gas", "in", t] for t in timesteps) * gasflow_co2



from typing import Any, Optional

import pandas as pd

from oogeso import dto


def compute_kpis(
    sim_result: dto.SimulationResult,
    sim_data: dto.EnergySystemData,
    windturbines: Optional[Any] = None,  # Fixme: What is correct type here?
):
    """Compute key indicators of simulation results

    sim_result :

    """
    hour_per_year = 8760
    sec_per_year = 3600 * hour_per_year
    kpi = {}
    res = sim_result
    if windturbines is None:
        windturbines = []

    num_sim_timesteps = res.co2_rate.shape[0]
    timesteps = res.co2_rate.index
    td_min = sim_data.parameters.time_delta_minutes
    kpi["hours_simulated"] = num_sim_timesteps * td_min / 60

    # CO2 emissions
    kpi["kgCO2_per_year"] = res.co2_rate.mean() * sec_per_year
    kpi["kgCO2_per_Sm3oe"] = res.co2_intensity.mean()

    # hours with reduced load
    #        kpi['reducedload_hours_per_year'] = None

    # hours with load shedding
    #        kpi['loadshed_hours_per_year'] = None

    # fuel consumption
    gasturbines = [
        g.id
        for g in sim_data.devices
        # if isinstance(g, Gasturbine) # why doesn't isinstance work?
        if g.model == "gasturbine"
    ]
    mask_gt = res.device_flow.index.get_level_values("device").isin(gasturbines)
    gtflow = res.device_flow[mask_gt]
    fuel = gtflow.unstack("carrier")["gas"].unstack("terminal")["in"].unstack().mean(axis=1)
    kpi["gt_fuel_sm3_per_year"] = fuel.sum() * sec_per_year

    # electric power consumption
    el_dem = res.device_flow.unstack("carrier")["el"].unstack("terminal")["in"].dropna().unstack().mean(axis=1)
    kpi["elconsumption_mwh_per_year"] = el_dem.sum() * hour_per_year
    kpi["elconsumption_avg_mw"] = el_dem.sum()

    # number of generator starts
    gt_starts = res.device_starting.unstack().sum(axis=1)[gasturbines].sum()
    kpi["gt_starts_per_year"] = gt_starts * hour_per_year / kpi["hours_simulated"]

    # number of generator stops
    gt_stops = res.device_stopping.unstack().sum(axis=1)[gasturbines].sum()
    kpi["gt_stops_per_year"] = gt_stops * hour_per_year / kpi["hours_simulated"]

    # running hours of generators
    gt_ison_tsteps = res.device_is_on.unstack().sum(axis=1)[gasturbines].sum()
    gt_ison = gt_ison_tsteps * td_min / 60
    kpi["gt_hoursrunning_per_year"] = gt_ison * hour_per_year / kpi["hours_simulated"]

    # wind power output
    el_sup = res.device_flow.unstack("carrier")["el"].unstack("terminal")["out"].dropna().unstack()
    p_wind = el_sup.T[windturbines]
    kpi["wind_output_mwh_per_year"] = p_wind.sum(axis=1).mean() * hour_per_year

    # curtailed wind energy
    p_avail = pd.DataFrame(index=timesteps)
    for devdata in sim_data.devices:
        if devdata in windturbines:
            d = devdata.id
            Pmax = devdata.flow_max
            p_avail[d] = Pmax
            if devdata.profile is not None:
                profile_ref = devdata.profile
                p_avail[d] = Pmax * res.profiles_nowcast.loc[timesteps, profile_ref]
    p_curtailed = (p_avail - p_wind).sum(axis=1)
    kpi["wind_curtailed_mwh_per_year"] = p_curtailed.mean() * hour_per_year
    return kpi



import typing

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.networks.edge import Edge


class Network:
    def __init__(
        self,
        carrier_data: dto.CarrierData,
        edges: typing.Dict[str, Edge],
    ):
        """Energy carrier."""
        self.carrier_id = carrier_data.id
        self.carrier_data = carrier_data
        self.edges = edges

    def define_constraints(self, pyomo_model: pyo.Model):
        piecewise_repn = pyo.value(pyomo_model.paramPiecewiseRepn)
        for edge in self.edges.values():
            edge.define_constraints(pyomo_model=pyomo_model, piecewise_repn=piecewise_repn)


class Heat(Network):
    pass


class Hydrogen(Network):
    pass



import logging
from typing import Dict, Union

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.devices.base import Device
from oogeso.core.networks.network_node import NetworkNode

logger = logging.getLogger(__name__)


class PumpDevice(Device):
    """Parent class for pumps. Don't use this class directly."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def compute_pump_demand(self, pyomo_model: pyo.Model, carrier, linear=False, Q=None, p1=None, p2=None, t=None):
        # Fixne: Fix types in arguments above.
        dev_data = self.dev_data
        node_id = dev_data.node_id
        node_obj: NetworkNode = self.node
        # power demand vs flow rate and pressure difference
        # see eg. doi:10.1016/S0262-1762(07)70434-0
        # P = Q*(p_out-p_in)/eta
        # units: m3/s*MPa = MW
        #
        # Assuming incompressible fluid, so flow rate m3/s=Sm3/s
        # (this approximation may not be very good for multiphase
        # wellstream)
        # assume nominal pressure and keep only flow rate dependence
        # TODO: Better linearisation?

        eta = dev_data.eta

        if t is None:
            t = 0
        if Q is None:
            Q = pyomo_model.varDeviceFlow[self.id, carrier, "in", t]
        if p1 is None:
            p1 = pyomo_model.varPressure[node_id, carrier, "in", t]
        if p2 is None:
            p2 = pyomo_model.varPressure[node_id, carrier, "out", t]
        if linear:
            # linearised equations around operating point
            # p1=p10, p2=p20, Q=Q0
            if t == 0:
                logger.debug("Node:{}, nominal pressures={}".format(node_id, node_obj.pressure_nominal))
            p10 = node_obj.pressure_nominal[carrier]["in"]
            p20 = node_obj.pressure_nominal[carrier]["out"]
            delta_p = p20 - p10
            #            Q0 = model.paramDevice[dev]['Q0']
            # P = (Q*(p20-p10)+Q0*(p10-p1))/eta
            P = Q * delta_p / eta
        #        elif self._quadraticConstraints:
        #            # Quadratic constraint...
        #            delta_p = (model.varPressure[(node,carrier,'out',t)]
        #                        -model.varPressure[(node,carrier,'in',t)])
        else:
            # non-linear equation - for computing outside optimisation
            P = Q * (p2 - p1) / eta
        return P

    def _rules_pump(self, model: pyo.Model, t: int, i: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        dev = self.id
        if isinstance(self, PumpOil):
            carrier = "oil"
        elif isinstance(self, PumpWater):
            carrier = "water"
        else:
            raise NotImplementedError(f"Device model {self.dev_data.model} has not been implemented")
        if i == 1:
            # flow out = flow in
            lhs = model.varDeviceFlow[dev, carrier, "out", t]
            rhs = model.varDeviceFlow[dev, carrier, "in", t]
            return lhs == rhs
        elif i == 2:
            lhs = model.varDeviceFlow[dev, "el", "in", t]
            rhs = self.compute_pump_demand(
                pyomo_model=model,
                carrier=carrier,
                linear=True,
                t=t,
            )
            return lhs == rhs
        else:
            raise ValueError(f"Argument i must be 1 or 2. {i} was given.")

    def define_constraints(self, pyomo_model: pyo.Model):
        """Specifies the list of constraints for the device"""

        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr = pyo.Constraint(pyomo_model.setHorizon, pyo.RangeSet(1, 2), rule=self._rules_pump)
        # add constraint to model:
        setattr(pyomo_model, "constr_{}_{}".format(self.id, "misc"), constr)
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "el", "in", t]


class PumpOil(PumpDevice):
    """Oil pump"""

    carrier_in = ["oil", "el"]
    carrier_out = ["oil"]
    serial = ["oil"]

    def __init__(
        self,
        dev_data: dto.DevicePumpOilData,  # Fixme: Correct?
        carrier_data_dict: Dict[str, Union[dto.CarrierElData, dto.CarrierOilData]],  # Fixme: Correct?
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict


class PumpWellStream(PumpDevice):
    """Wellstream pump"""

    carrier_in = ["wellstream", "el"]
    carrier_out = ["wellstream"]
    serial = ["wellstream"]

    # Fixme: Missing DTO


class PumpWater(PumpDevice):
    """Water pump"""

    carrier_in = ["water", "el"]
    carrier_out = ["water"]
    serial = ["water"]

    def __init__(
        self,
        dev_data: dto.DevicePumpWaterData,  # Fixme: Correct?
        carrier_data_dict: Dict[str, Union[dto.CarrierWaterData, dto.CarrierElData]],  # Fixme: Correct?
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict



from .oogeso_input_data_objects import (
    CarrierData,
    CarrierElData,
    DeviceData,
    DevicePowerSinkData,
    DevicePowerSourceData,
    DeviceStorageElData,
    EdgeData,
    EdgeElData,
    EnergySystemData,
    NodeData,
    OptimisationParametersData,
    StartStopData,
    TimeSeriesData,
)
from .oogeso_input_data_objects_multienergy import (
    CarrierGasData,
    CarrierHeatData,
    CarrierHydrogenData,
    CarrierOilData,
    CarrierWaterData,
    CarrierWellstreamData,
    DeviceCompressorElData,
    DeviceCompressorGasData,
    DeviceElectrolyserData,
    DeviceFuelCellData,
    DeviceGasHeaterData,
    DeviceGasTurbineData,
    DeviceHeatPumpData,
    DevicePumpOilData,
    DevicePumpWaterData,
    DeviceSeparator2Data,
    DeviceSeparatorData,
    DeviceSinkElData,
    DeviceSinkGasData,
    DeviceSinkHeatData,
    DeviceSinkOilData,
    DeviceSinkWaterData,
    DeviceSourceElData,
    DeviceSourceGasData,
    DeviceSourceWaterData,
    DeviceStorageHydrogenData,
    DeviceWellGasLiftData,
    DeviceWellProductionData,
    EdgeFluidData,
    EdgeGasData,
    EdgeHeatData,
    EdgeHydrogenData,
    EdgeOilData,
    EdgeWaterData,
    EdgeWellstreamData,
)
from .oogeso_output_data_objects import SimulationResult



from typing import Optional, Union

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.networks.network_node import NetworkNode


class Edge:
    """Network edge"""

    def __init__(self, edge_data_object: dto.EdgeData):
        self.id = edge_data_object.id
        self.edge_data = edge_data_object  # Edge data object as defined in the DTO
        self.edges = {}
        self.node_from: Optional[NetworkNode] = None
        self.node_to: Optional[NetworkNode] = None

    def add_nodes(self, node_from: NetworkNode, node_to: NetworkNode):
        self.node_from = node_from
        self.node_to = node_to

    def define_constraints(self, pyomo_model: pyo.Model, piecewise_repn):
        """Builds constraints for the edge"""

        constr_edge_bounds = pyo.Constraint(pyomo_model.setHorizon, rule=self._rule_edge_flow_max_min)
        setattr(
            pyomo_model,
            "constrE_{}_{}".format(self.id, "bounds"),
            constr_edge_bounds,
        )

        # Losses (flow out of edge less than into edge)
        if self.has_loss():

            # First, connecting variables (flow in different directions and loss variables)
            constr_loss = pyo.Constraint(
                pyomo_model.setHorizon,
                pyo.RangeSet(1, 2),
                rule=self._rule_edge_flow_and_loss,
            )
            setattr(pyomo_model, "constrE_{}_{}".format(self.id, "loss"), constr_loss)
            # Then, add equations for losses vs power flow (piecewise linear equations):
            for i in [1, 2]:
                constr_loss_function = self._loss_function_constraint(i, pyomo_model, piecewise_repn)
                setattr(
                    pyomo_model,
                    "constrE_{}_{}_{}".format(self.id, "lossfunction", i),
                    constr_loss_function,
                )

    def _rule_edge_flow_max_min(self, pyomo_model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        edge = self.id
        # if non-bidirectional, lower bound=0, if bidirectional, bound=+- pmax
        if self.edge_data.flow_max is None:
            if self.edge_data.bidirectional:
                # unconstrained
                expr = pyo.Constraint.Skip
            else:
                # one-directional, so lower bound 0 (no upper bound specified)
                # expr = pyo.inequality(0, model.varEdgeFlow[edge, t], None)
                expr = pyomo_model.varEdgeFlow[edge, t] >= 0
        else:
            pmax = self.edge_data.flow_max
            if self.edge_data.bidirectional:
                expr = pyo.inequality(-pmax, pyomo_model.varEdgeFlow[edge, t], pmax)
            else:
                expr = pyo.inequality(0, pyomo_model.varEdgeFlow[edge, t], pmax)
        return expr

    def has_loss(self) -> bool:
        # Fixme: Implement power_loss_function as optional in EdgeData
        has_loss = hasattr(self.edge_data, "power_loss_function") and (self.edge_data.power_loss_function is not None)
        return has_loss

    def _rule_edge_flow_and_loss(
        self, pyomo_model: pyo.Model, t: int, i: int
    ) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """Split edge flow into positive and negative part, for loss calculations"""
        edge = self.id
        if i == 1:
            expr = (
                pyomo_model.varEdgeFlow[edge, t]
                == pyomo_model.varEdgeFlow12[edge, t] - pyomo_model.varEdgeFlow21[edge, t]
            )
        elif i == 2:
            expr = (
                pyomo_model.varEdgeLoss[edge, t]
                == pyomo_model.varEdgeLoss12[edge, t] + pyomo_model.varEdgeLoss21[edge, t]
            )
        else:
            raise NotImplementedError("Only i 1 and 2 is implemented.")
        return expr

    def _loss_function_constraint(self, i: int, pyomo_model: pyo.Model, piecewise_repn):

        # Piecewise constraints require independent variable to be bounded:
        pyomo_model.varEdgeFlow12[self.id, :].setub(self.edge_data.flow_max)
        pyomo_model.varEdgeFlow21[self.id, :].setub(self.edge_data.flow_max)
        # Losses on cables are: P_loss = R/V^2 * P^2, i.e. quadratic function of power flow
        # Losses in transformers are: P_loss = ...
        lookup_table = self.edge_data.power_loss_function
        pw_x = lookup_table[0]
        pw_y = lookup_table[1]
        # # If we use loss function giving loss fraction instead of absolute loss in MW:
        # pw_y_fraction = lookup_table[1]
        # # Lookup table gives losses as a fraction, so to get absolute values we need
        # # to multiply by power transfer
        # # NOTE: This changes a linear curve to a second order curve, so need more than
        # # two x-points to represent it properly.
        # pw_y = [pw_y_fraction[i] * pw_x[i] for i in len(pw_x)]
        if i == 1:
            var_x = pyomo_model.varEdgeFlow12
            var_y = pyomo_model.varEdgeLoss12
        elif i == 2:
            var_x = pyomo_model.varEdgeFlow21
            var_y = pyomo_model.varEdgeLoss21
        else:
            raise NotImplementedError("i other than (1, 2) has not been implemented.")
        constr_penalty = pyo.Piecewise(
            [self.id],
            pyomo_model.setHorizon,
            var_y,
            var_x,
            pw_repn=piecewise_repn,
            pw_constr_type="EQ",
            pw_pts=pw_x,
            f_rule=pw_y,
        )
        return constr_penalty



import re
from dataclasses import dataclass, field
from typing import List, Optional, Tuple, Union


@dataclass
class NodeData:
    # unique identifier:
    id: str


@dataclass
class StartStopData:
    is_on_init: bool = False  # Initial on/off status
    penalty_start: float = 0  # Startup "cost"
    penalty_stop: float = 0  # Shutdown "cost"
    delay_start_minutes: int = 0  # Delay in minutes from activation to online
    minimum_time_on_minutes: float = 0  # Minimum on-time in minutes once started
    minimum_time_off_minutes: float = 0  # Minimum off-time in minutes once stopped


@dataclass
class DeviceData:  # Parent class - use subclasses instead
    id: str  # unique identifier
    node_id: str  # reference to node identifier
    name: str = ""
    include: Optional[bool] = True
    profile: Optional[str] = None  # reference to time-series
    flow_min: Optional[float] = None  # Energy or fluid flow limit
    flow_max: Optional[float] = None
    # Ramp rates are given as change relative to capacity per minute, 1=100%/min:
    max_ramp_down: Optional[float] = None
    max_ramp_up: Optional[float] = None
    start_stop: Optional[StartStopData] = None
    reserve_factor: float = 0  # contribution to electrical spinning reserve
    op_cost: Optional[float] = None
    model: str = field(init=False)  # model name is derived from class name

    def __post_init__(self):
        # Model name is given by class name. Eg "gasturbine" <-> "DeviceGasturbineData"
        # This field is added to know (when serialized) which subclass the device belongs to
        modelname = re.search("Device(.+?)Data", self.__class__.__name__).group(1)
        self.model = modelname.lower()


@dataclass
class DevicePowerSourceData(DeviceData):
    # Penalty may be fuel, emissions, cost and combinations of these
    penalty_function: Tuple[List[float], List[float]] = None
    reserve_factor: float = 1  # not used capacity contributes fully to spinning reserve


@dataclass
class DevicePowerSinkData(DeviceData):
    pass


@dataclass
class DeviceStorageElData(DeviceData):
    E_max: float = 0  # MWh storage capacity
    E_min: float = 0
    E_end: Optional[float] = None  # required storage level at end of horzion
    E_cost: Optional[float] = None  # cost for non-ful storage
    eta: float = 1  # efficiency
    target_profile: Optional[str] = None
    E_init: Optional[float] = 0


@dataclass
class EdgeData:
    id: str
    node_from: str
    node_to: str
    length_km: Optional[float] = None
    flow_max: float = None  # Maximum flow (MW or Sm3/s)
    bidirectional: Optional[bool] = True
    include: bool = True  # whether to include object in problem formulation
    carrier: str = field(init=False)

    def __post_init__(self):
        carrier_name = re.search("Edge(.+?)Data", self.__class__.__name__).group(1)
        self.carrier = carrier_name.lower()


@dataclass
class EdgeElData(EdgeData):
    resistance: float = 0  # ohm per km
    reactance: float = 0  # ohm per km
    # Voltage for line (single value) or transformer (tuple)
    voltage: Union[float, Tuple[float, float]] = None  # kV.
    # Power loss in MW as function of power transfer in MW:
    power_loss_function: Optional[Tuple[List[float], List[float]]] = None


@dataclass
class CarrierData:
    id: str


@dataclass
class CarrierElData(CarrierData):
    powerflow_method: str = "transport"  # "transport","dc_pf"
    reference_node: str = None  # reference node for dc-pf electrical voltage angles
    # required (globally) spinning reserve (MW), -1=no limit
    el_reserve_margin: float = -1
    # required backup margin (MW), -1=no limit
    el_backup_margin: Optional[float] = -1  # MW, -1=no limit


@dataclass
class OptimisationParametersData:
    # name of objective function to use:
    objective: str
    # minutes per timestep:
    time_delta_minutes: int
    # timesteps in each rolling optimisation:
    planning_horizon: int
    # timesteps between each optimisation:
    optimisation_timesteps: int
    # timesteps beyond which forecast (instead of nowcast) profile is used:
    forecast_timesteps: int
    # minutes, how long stored energy must be sustained to count as reserve:
    time_reserve_minutes: Optional[int] = None
    # costs for co2 emissions (currency/kgCO2)
    co2_tax: Optional[float] = None
    # global limit for allowable relative pressure deviation from nominal:
    max_pressure_deviation: float = -1
    # limit on allowable emission intensity (kgCO2/Sm3oe), -1=no limit
    emission_intensity_max: Optional[float] = -1
    # limit on allowable emission intensity (kgCO2/hour), -1= no limit
    emission_rate_max: Optional[float] = -1
    # how to represent piecewise linear constraints:
    piecewise_repn: str = "SOS2"
    # specify which data to return from simulation as a list. None=return all data
    optimisation_return_data: Optional[List[str]] = None


TimeSeries = List[float]


@dataclass
class TimeSeriesData:
    id: str
    data: TimeSeries
    data_nowcast: Optional[TimeSeries] = None


@dataclass
class EnergySystemData:
    parameters: OptimisationParametersData
    carriers: List[CarrierData]
    nodes: List[NodeData]
    edges: List[EdgeData]
    # devices: Dict[str, DeviceData]
    devices: List[DeviceData]
    profiles: List[TimeSeriesData]
    # profiles: Dict[str, TimeSeries]



from typing import Dict, Union

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.devices.base import Device


class HeatPump(Device):
    """
    Heat pump or electric heater (el to heat)
    """

    carrier_in = ["el"]
    carrier_out = ["heat"]
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceHeatPumpData,  # Fixme: Correct?
        carrier_data_dict: Dict[str, dto.CarrierElData],  # Fixme: Correct?
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def _rules(self, pyomo_model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        dev = self.id
        # heat out = el in * efficiency
        lhs = pyomo_model.varDeviceFlow[dev, "heat", "out", t]
        rhs = pyomo_model.varDeviceFlow[dev, "el", "in", t] * pyomo_model.paramDevice[dev]["eta"]
        return pyo.Expression(lhs == rhs)

    def define_constraints(self, pyomo_model: pyo.Model):
        """Specifies the list of constraints for the device"""

        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr = pyo.Constraint(pyomo_model.setHorizon, rule=self._rules)
        # add constraint to model:
        setattr(pyomo_model, "constr_{}_{}".format(self.id, "misc"), constr)
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "el", "in", t]



from .compressor import CompressorEl, CompressorGas
from .electrolyser import Electrolyser
from .fuelcell import FuelCell
from .gasheater import GasHeater
from .gasturbine import GasTurbine
from .heatpump import HeatPump
from .pump import PumpOil, PumpWater, PumpWellStream
from .separator import Separator, Separator2
from .sink import PowerSink, SinkEl, SinkGas, SinkHeat, SinkOil, SinkWater
from .source import Powersource, SourceEl, SourceGas, SourceOil, SourceWater
from .storage import StorageEl, StorageHydrogen
from .well import WellGasLift, WellProduction



from typing import Dict

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.devices.base import Device


class PowerSink(Device):
    """
    Generic electricity consumption
    """

    carrier_in = ["el"]
    carrier_out = []
    serial = []

    def __init__(
        self,
        dev_data: dto.DevicePowerSinkData,
        carrier_data_dict: Dict[str, dto.CarrierElData],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "el", "in", t]


# Just another name for powersink
class SinkEl(PowerSink):
    pass


class SinkHeat(Device):
    "Generic heat consumption"
    carrier_in = ["heat"]
    carrier_out = []
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceSinkHeatData,
        carrier_data_dict: Dict[str, dto.CarrierHeatData],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "heat", "in", t]


class SinkGas(Device):
    "Generic electricity consumption"
    carrier_in = ["gas"]
    carrier_out = []
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceSinkGasData,
        carrier_data_dict: Dict[str, dto.CarrierGasData],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "gas", "in", t]


class SinkOil(Device):
    "Generic oil consumption"
    carrier_in = ["oil"]
    carrier_out = []
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceSinkOilData,
        carrier_data_dict: Dict[str, dto.CarrierOilData],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "oil", "in", t]


class SinkWater(Device):
    "Generic water consumption"
    carrier_in = ["water"]
    carrier_out = []
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceSinkWaterData,
        carrier_data_dict: Dict[str, dto.CarrierWaterData],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def rule_devmodel_sink_water(self, model: pyo.Model, t: int, i: int):
        dev = self.id
        dev_data = self.dev_data
        time_delta_minutes = model.paramTimestepDeltaMinutes

        if dev_data.flow_avg is None:
            return pyo.Constraint.Skip
        if dev_data.max_accumulated_deviation is None:
            return pyo.Constraint.Skip
        if dev_data.max_accumulated_deviation == 0:
            return pyo.Constraint.Skip
        if i == 1:
            # FLEXIBILITY
            # (water_in-water_avg)*dt = delta buffer
            delta_t = time_delta_minutes / 60  # hours
            lhs = (model.varDeviceFlow[dev, "water", "in", t] - dev_data.flow_avg) * delta_t
            if t > 0:
                Eprev = model.varDeviceStorageEnergy[dev, t - 1]
            else:
                Eprev = model.paramDeviceEnergyInitially[dev]
            rhs = model.varDeviceStorageEnergy[dev, t] - Eprev
            return lhs == rhs
        elif i == 2:
            # energy buffer limit
            Emax = dev_data.max_accumulated_deviation
            return pyo.inequality(-Emax / 2, model.varDeviceStorageEnergy[dev, t], Emax / 2)
        else:
            raise Exception("impossible")

    def define_constraints(self, pyomo_model: pyo.Model):
        """Specifies the list of constraints for the device"""
        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr = pyo.Constraint(
            pyomo_model.setHorizon,
            pyo.RangeSet(1, 2),
            rule=self.rule_devmodel_sink_water,
        )
        # add constraints to model:
        setattr(pyomo_model, "constr_{}_{}".format(self.id, "flex"), constr)
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "water", "in", t]



from typing import Dict, List, Union

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.devices.base import Device


class SourceEl(Device):
    "Generic external source for electricity (e.g. cable or wind turbine)"
    carrier_in = []
    carrier_out = ["el"]
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceSourceElData,
        carrier_data_dict: Dict[str, dto.CarrierElData],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "el", "out", t]

    # overriding default
    def compute_CO2(self, pyomo_model: pyo.Model, timesteps: List[int]):
        # co2 content in fuel combustion
        # co2em is kgCO2/MWh_el, deltaT is seconds, deviceFlow is MW
        # need to convert co2em to kgCO2/(MW*s)
        thisCO2 = 0
        if self.dev_data.co2em is not None:
            thisCO2 = (
                sum(pyomo_model.varDeviceFlow[self.id, "el", "out", t] * self.dev_data.co2em for t in timesteps)
                * 1
                / 3600
            )
        return thisCO2


class Powersource(Device):
    """Generic external source for electricity (e.g. cable or wind turbine)."""

    carrier_in = []
    carrier_out = ["el"]
    serial = []

    def __init__(
        self,
        dev_data: dto.DevicePowerSourceData,
        carrier_data_dict: Dict[str, dto.CarrierElData],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def define_constraints(self, pyomo_model: pyo.Model):
        """Specifies the list of constraints for the device"""

        list_to_reconstruct = super().define_constraints(pyomo_model)

        if self.dev_data.penalty_function is not None:
            constr_penalty = self._penaltyConstraint(pyomo_model)
            setattr(
                pyomo_model,
                "constrPW_{}_{}".format(self.id, "penalty"),
                constr_penalty,
            )
        return list_to_reconstruct

    # TODO: Make piecewise constraint implementation more elegant
    # The varDevicePenalty seem to be required to have the same indices as the varDeviceFlow.
    # But only (device_id, time) is relevant
    # Could define a "varDevicePower" (or better name) representing the main variable
    # used with p_max / q_max / penalty_function
    def _penaltyConstraint(self, pyomo_model: pyo.Model):
        # Piecewise constraints require independent variable to be bounded:
        # ub = self.dev_data.flow_max
        ub = self.get_flow_upper_bound()
        pyomo_model.varDeviceFlow[self.id, "el", "out", :].setlb(0)
        pyomo_model.varDeviceFlow[self.id, "el", "out", :].setub(ub)
        lookup_table = self.dev_data.penalty_function
        pw_x = lookup_table[0]
        pw_y = lookup_table[1]
        var_x = pyomo_model.varDeviceFlow  # [self.dev_id, "el", "out", :]
        var_y = pyomo_model.varDevicePenalty  # [self.dev_id, "el", "out", :]
        pw_repn = pyo.value(pyomo_model.paramPiecewiseRepn)
        constr_penalty = pyo.Piecewise(
            [self.id],
            ["el"],
            ["out"],
            pyomo_model.setHorizon,
            var_y,
            var_x,
            pw_repn=pw_repn,  # default "SOS2" does not work with CBC solver
            pw_constr_type="EQ",
            pw_pts=pw_x,
            f_rule=pw_y,  # self._penaltyfunction,
        )
        return constr_penalty

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "el", "out", t]


class SourceGas(Device):
    """Generic external source for gas."""

    carrier_in = []
    carrier_out = ["gas"]
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceSourceGasData,
        carrier_data_dict: Dict[str, dto.CarrierGasData],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def _rules(self, model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        node = self.dev_data.node_id
        lhs = model.varPressure[(node, "gas", "out", t)]
        rhs = self.dev_data.naturalpressure
        return lhs == rhs

    def define_constraints(self, pyomo_model: pyo.Model):
        """Specifies the list of constraints for the device"""
        # add generic device constraints:
        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr_well = pyo.Constraint(pyomo_model.setHorizon, rule=self._rules)
        # add constraint to model:
        setattr(
            pyomo_model,
            "constr_{}_{}".format(self.id, "pressure"),
            constr_well,
        )
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "gas", "out", t]


class SourceOil(Device):
    """Generic external source for oil."""

    carrier_in = []
    carrier_out = ["oil"]
    serial = []

    # Fixme: Missing DTOs?

    def _rules(self, model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        node = self.dev_data.node_id
        lhs = model.varPressure[(node, "oil", "out", t)]
        rhs = self.dev_data.naturalpressure
        return lhs == rhs

    def define_constraints(self, pyomo_model: pyo.Model):
        """Specifies the list of constraints for the device"""
        # add generic device constraints:
        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr_well = pyo.Constraint(pyomo_model.setHorizon, rule=self._rules)
        # add constraint to model:
        setattr(
            pyomo_model,
            "constr_{}_{}".format(self.id, "pressure"),
            constr_well,
        )
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "oil", "out", t]


class SourceWater(Device):
    """Generic external source for water."""

    carrier_in = []
    carrier_out = ["water"]
    serial = []

    def __init__(
        self,
        dev_data: dto.DeviceSourceWaterData,
        carrier_data_dict: Dict[str, dto.CarrierWaterData],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def _rules(self, model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        node = self.dev_data.node_id
        lhs = model.varPressure[(node, "water", "out", t)]
        rhs = self.dev_data.naturalpressure
        return lhs == rhs

    def define_constraints(self, pyomo_model: pyo.Model):
        """Specifies the list of constraints for the device"""
        # add generic device constraints:
        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr_well = pyo.Constraint(pyomo_model.setHorizon, rule=self._rules)
        # add constraint to model:
        setattr(
            pyomo_model,
            "constr_{}_{}".format(self.id, "pressure"),
            constr_well,
        )
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int):
        return pyomo_model.varDeviceFlow[self.id, "water", "out", t]



from typing import Dict, List, Optional, Union

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.devices.base import Device
from oogeso.core.networks.network_node import NetworkNode


class CompressorEl(Device):
    """Electric compressor."""

    carrier_in = ["gas", "el"]
    carrier_out = ["gas"]
    serial = ["gas"]

    def __init__(
        self,
        dev_data: Union[dto.DeviceCompressorGasData, dto.DeviceCompressorElData],
        carrier_data_dict: Dict[str, Union[dto.CarrierElData, dto.CarrierGasData]],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def _rules(self, pyomo_model: pyo.Model, t: int, i: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        dev = self.id
        node_obj: NetworkNode = self.node
        gas_data = self.carrier_data["gas"]
        if i == 1:  # Fixme: use carrier_data object type instead?
            """gas flow in equals gas flow out (mass flow)"""
            lhs = pyomo_model.varDeviceFlow[dev, "gas", "in", t]
            rhs = pyomo_model.varDeviceFlow[dev, "gas", "out", t]
            return lhs == rhs
        elif i == 2:  # Fixme: use carrier_data object type instead?
            """Device el demand"""
            lhs = pyomo_model.varDeviceFlow[dev, "el", "in", t]
            rhs = compute_compressor_demand(pyomo_model, self, node_obj, gas_data, linear=True, t=t)
            return lhs == rhs

    def define_constraints(self, pyomo_model: pyo.Model) -> List[pyo.Constraint]:
        """Specifies the list of constraints for the device"""

        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr_compressor_el = pyo.Constraint(pyomo_model.setHorizon, pyo.RangeSet(1, 2), rule=self._rules)
        # add constraint to model:
        setattr(
            pyomo_model,
            "constr_{}_{}".format(self.id, "compr"),
            constr_compressor_el,
        )
        return list_to_reconstruct

    def get_flow_var(self, pyomo_model: pyo.Model, t: int) -> pyo.Var:
        return pyomo_model.varDeviceFlow[self.id, "el", "in", t]


class CompressorGas(Device):
    """
    Gas-driven compressor
    """

    carrier_in = ["gas"]
    carrier_out = ["gas"]
    serial = ["gas"]

    def __init__(
        self,
        dev_data: dto.DeviceCompressorGasData,
        carrier_data_dict: Dict[str, dto.CarrierGasData],
    ):
        super().__init__(dev_data=dev_data, carrier_data_dict=carrier_data_dict)
        self.dev_data = dev_data
        self.id = dev_data.id
        self.carrier_data = carrier_data_dict

    def _rules(self, pyomo_model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        dev = self.id
        dev_data: DeviceCompressorGasData = self.dev_data  # noqa: Fixme: dev_data not in use.
        node_obj: NetworkNode = self.node
        gas_data = self.carrier_data["gas"]
        gas_energy_content = gas_data.energy_value  # MJ/Sm3
        power_demand = compute_compressor_demand(pyomo_model, self, node_obj, gas_data, linear=True, t=t)

        # matter conservation:
        lhs = pyomo_model.varDeviceFlow[dev, "gas", "out", t]
        rhs = pyomo_model.varDeviceFlow[dev, "gas", "in", t] - power_demand / gas_energy_content

        return lhs == rhs

    def define_constraints(self, pyomo_model: pyo.Model):
        """Specifies the list of constraints for the device"""

        list_to_reconstruct = super().define_constraints(pyomo_model)

        constr = pyo.Constraint(pyomo_model.setHorizon, rule=self._rules)
        setattr(pyomo_model, "constr_{}_{}".format(self.id, "compr"), constr)
        return list_to_reconstruct

    # overriding default
    def compute_CO2(self, pyomo_model: pyo.Model, timesteps: List[int]):
        d = self.id
        gas_data = self.carrier_data["gas"]
        gasflow_co2 = gas_data.co2_content  # kg/m3

        return (
            sum(
                (pyomo_model.varDeviceFlow[d, "gas", "in", t] - pyomo_model.varDeviceFlow[d, "gas", "out", t])
                for t in timesteps
            )
            * gasflow_co2
        )

    def get_flow_var(self, pyomo_model: pyo.Model, t: int) -> float:
        raise NotImplementedError()


def compute_compressor_demand(
    model,
    device_obj: Union[CompressorEl, CompressorGas],
    node_obj: NetworkNode,
    gas_data: dto.CarrierGasData,
    linear: bool = False,
    Q: Optional[float] = None,
    p1: Optional[float] = None,
    p2: Optional[float] = None,
    t: Optional[float] = None,
):
    """Compute energy demand by compressor as function of pressure and flow"""
    # power demand depends on gas pressure ratio and flow
    # See LowEmission report DSP5_2020_04 for description

    dev_data = device_obj.dev_data
    k = gas_data.k_heat_capacity_ratio
    Z = gas_data.Z_compressibility
    # factor 1e-6 converts R units from J/kgK to MJ/kgK:
    R = gas_data.R_individual_gas_constant * 1e-6
    rho = gas_data.rho_density
    T1 = dev_data.temp_in  # inlet temperature
    eta = dev_data.eta  # isentropic efficiency
    a = (k - 1) / k
    c = rho / eta * 1 / (k - 1) * Z * R * T1
    node = node_obj.id
    if t is None:
        t = 0
    if Q is None:
        Q = model.varDeviceFlow[device_obj.id, "gas", "out", t]
    if p1 is None:
        p1 = model.varPressure[node, "gas", "in", t]
    if p2 is None:
        p2 = model.varPressure[node, "gas", "out", t]
    if linear:
        # linearised equations around operating point
        # p1=p10, p2=p20, Q=Q0
        p10 = node_obj.pressure_nominal["gas"]["in"]
        p20 = node_obj.pressure_nominal["gas"]["out"]
        Q0 = dev_data.Q0
        P = c * (a * (p20 / p10) ** a * Q0 * (p2 / p20 - p1 / p10) + ((p20 / p10) ** a - 1) * Q)
    else:
        P = c * ((p2 / p1) ** a - 1) * Q
    return P



import logging
from typing import Dict, Union

import pandas as pd
import pyomo.environ as pyo
import pyomo.opt as pyopt

from oogeso import dto
from oogeso.core import networks
from oogeso.core.devices.base import Device
from oogeso.core.devices.storage import StorageDevice
from oogeso.core.networks.edge import Edge
from oogeso.core.networks.network import Network
from oogeso.core.networks.network_node import NetworkNode
from oogeso.utils.util import get_device_from_model_name

logger = logging.getLogger(__name__)


class OptimisationModel(pyo.ConcreteModel):
    """Class for MILP optimisation model"""

    ZERO_WARNING_THRESHOLD = 1e-6

    def __init__(self, data: dto.EnergySystemData):
        """Create optimisation problem formulation with supplied data"""

        super().__init__()

        # dictionaries {key:object} for all devices, nodes and edges
        self.all_devices: Dict[str, Device] = {}
        self.all_nodes: Dict[str, NetworkNode] = {}
        self.all_edges: Dict[str, Edge] = {}
        # self.all_carriers = {}
        self.all_networks: Dict[str, Network] = {}

        # Model parameters
        self.optimisation_parameters = data.parameters

        # List of constraints that need to be reconstructed for each optimisation:
        self.constraints_to_reconstruct = []
        # List of devices with storage
        self.devices_with_storage = []
        profiles_in_use = list(set(d.profile for d in data.devices if d.profile is not None))
        logger.info("profiles in use: %s", profiles_in_use)

        self._create_network_objects_from_data(data)
        self._set_node_pressure_from_edge_data()
        self._create_pyomo_model(profiles_in_use)

    def solve(self, solver="cbc", solver_executable=None, solver_options=None, write_yaml=False, time_limit=None):
        """Solve problem for planning horizon at a single timestep"""

        opt = pyo.SolverFactory(solver, executable=solver_executable)
        if solver_options is not None:
            for k in solver_options:
                opt.options[k] = solver_options[k]
        if time_limit is not None:
            if solver == "gurobi":
                opt.options["TimeLimit"] = time_limit
            elif solver == "cbc":
                opt.options["sec"] = time_limit
            elif solver == "cplex":
                opt.options["timelimit"] = time_limit
            elif solver == "glpk":
                opt.options["tmlim"] = time_limit
        logger.debug("Solving...")
        sol = opt.solve(self)

        if write_yaml:
            sol.write_yaml()

        if (sol.solver.status == pyopt.SolverStatus.ok) and (
            sol.solver.termination_condition == pyopt.TerminationCondition.optimal
        ):
            logger.debug("Solved OK")
            pass
        elif sol.solver.termination_condition == pyopt.TerminationCondition.infeasible:
            raise Exception("Infeasible solution")
        else:
            # Something else is wrong
            logger.info("Solver Status:{}".format(sol.solver.status))
        return sol

    def _set_node_pressure_from_edge_data(self):
        # Set node terminal nominal pressure based on edge from/to pressure values
        # Fixme: hasattr is prone to bugs!
        for edge in self.all_edges.values():
            edg = edge.edge_data
            carrier = edg.carrier
            # Setting nominal pressure levels at node terminals. Raise exception
            # if inconsistencies are found
            if (hasattr(edg, "pressure_from")) and (edg.pressure_from is not None):
                n_from: NetworkNode = self.all_nodes[edg.node_from]
                p_from = edg.pressure_from
                n_from.set_pressure_nominal(carrier, "out", p_from)
            if (hasattr(edg, "pressure_to")) and (edg.pressure_to is not None):
                n_to: NetworkNode = self.all_nodes[edg.node_to]
                p_to = edg.pressure_to
                n_to.set_pressure_nominal(carrier, "in", p_to)
            # Setting max pressure deviation values at node terminals. Raise exception
            # if inconsistencies are found
            if (hasattr(edg, "pressure_from_maxdeviation")) and (edg.pressure_from_maxdeviation is not None):
                n_from: NetworkNode = self.all_nodes[edg.node_fom]
                p_maxdev_from = edg.pressure_from_maxdeviation
                n_from.set_pressure_maxdeviation(carrier, "out", p_maxdev_from)
            if (hasattr(edg, "pressure_to_maxdeviation")) and (edg.pressure_to_maxdeviation is not None):
                n_to: NetworkNode = self.all_nodes[edg.node_to]
                p_maxdev_to = edg.pressure_to_maxdeviation
                n_to.set_pressure_maxdeviation(carrier, "in", p_maxdev_to)

    def _create_network_objects_from_data(self, data: dto.EnergySystemData):
        """Create energy system objects, and populate local dictionaries

        self.all_devices, self.all_nodes, self.all_networks"""

        # json_str = json.dumps(data)
        # energy_system_data: inputdata.EnergySystemData = (
        #    inputdata.deserialize_oogeso_data(json_str)
        # )
        # energy_system_data = json.loads(json_str, cls=inputdata.DataclassJSONDecoder)
        energy_system_data = data

        # Create energy system network elements (devices, nodes, edges)
        for dev_data_obj in energy_system_data.devices:
            dev_id = dev_data_obj.id
            if not dev_data_obj.include:
                # skip this edge and move to next
                logger.debug("Excluding device {}".format(dev_id))
                continue
            device_model = dev_data_obj.model
            # The class corresponding to the device type should always have a
            # name identical to the type (but capitalized):
            logger.debug("Device model={}".format(device_model))
            device = get_device_from_model_name(model_name=device_model)
            carrier_data_dict = {carr_obj.id: carr_obj for carr_obj in data.carriers}
            new_device = device(dev_data_obj, carrier_data_dict)
            if isinstance(new_device, StorageDevice):
                # Add this device to global list of storage devices:
                self.devices_with_storage.append(new_device)

            new_device.set_flow_upper_bound(data.profiles)
            self.all_devices[dev_id] = new_device

        for node_data_obj in energy_system_data.nodes:
            new_node = networks.NetworkNode(node_data_obj)
            node_id = new_node.id
            self.all_nodes[node_id] = new_node

        edges_per_type = {}
        for edge_data_obj in data.edges:
            if edge_data_obj.include is False:
                # skip this edge and move to next
                continue
            edge_id = edge_data_obj.id
            carrier = edge_data_obj.carrier
            new_edge = networks.Edge(edge_data_obj)
            if carrier not in edges_per_type:
                edges_per_type[carrier] = {}
            edges_per_type[carrier][edge_id] = new_edge
            self.all_edges[edge_id] = new_edge

        for carrier_data_obj in data.carriers:
            carrier_model = carrier_data_obj.id  # el,heat,oil,gas,water,hydrogen,
            # The network class corresponding to the carrier should always have a
            # name identical to the id (but capitalized):
            network_class = getattr(networks, carrier_model.capitalize())
            if carrier_model not in edges_per_type:
                edges_per_type[carrier_model] = {}
            new_network = network_class(
                carrier_data=carrier_data_obj,
                edges=edges_per_type[carrier_model],
            )
            self.all_networks[carrier_model] = new_network

        # Make network connections between nodes, edges and devices
        for dev_id, dev in self.all_devices.items():
            logger.debug("Node-device: %s,%s", dev_id, dev)
            node_id_where_connected = dev.dev_data.node_id
            node = self.all_nodes[node_id_where_connected]
            node.add_device(dev_id, dev)
            dev.add_node(node)
        for edge_id, edge in self.all_edges.items():
            node_from = self.all_nodes[edge.edge_data.node_from]
            node_from.add_edge(edge, "from")
            node_to = self.all_nodes[edge.edge_data.node_to]
            node_to.add_edge(edge, "to")
            edge.add_nodes(node_from, node_to)

    def _create_pyomo_model(self, profiles_in_use):
        """Create pyomo MILP model

        Parameters:
        -----------
            planning_horizon: dict
            input data as read from input file
        """

        self._specify_sets_and_parameters(profiles_in_use)
        self._specify_variables()
        self._specify_objective()
        self._specify_constraints()

        # Specify initial values from input data
        for dev in self.all_devices.values():
            dev.set_init_values(pyomo_model=self)

        # Keep track of duals:
        # WARNING:
        # From Gurobi: "Shadow prices are not well-defined in mixed-integer
        # problems, so we don't provide dual values for an integer program."
        self.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)

    def _specify_sets_and_parameters(self, profiles_in_use):
        """specify pyomo model sets and parameters"""
        energycarriers = self.all_networks.keys()
        self.setCarrier = pyo.Set(initialize=energycarriers, doc="carrier")
        self.setTerminal = pyo.Set(initialize=["in", "out"], doc="terminal")
        self.setNode = pyo.Set(initialize=self.all_nodes.keys(), doc="node")
        self.setDevice = pyo.Set(initialize=self.all_devices.keys(), doc="device")
        self.setEdge = pyo.Set(initialize=self.all_edges.keys(), doc="edge")
        timerange = range(self.optimisation_parameters.planning_horizon)
        self.setHorizon = pyo.Set(ordered=True, initialize=timerange, doc="time")
        self.setProfile = pyo.Set(initialize=profiles_in_use, doc="profile")

        # specify mutable parameters:
        # (will be modified between successive optimisations)
        self.paramProfiles = pyo.Param(
            self.setProfile,
            self.setHorizon,
            within=pyo.Reals,
            mutable=True,
            initialize=1,
        )
        self.paramDeviceIsOnInitially = pyo.Param(self.setDevice, mutable=True, within=pyo.Binary, initialize=1)
        self.paramDevicePrepTimestepsInitially = pyo.Param(
            self.setDevice, mutable=True, within=pyo.Integers, initialize=0
        )
        # needed for ramp rate limits:
        self.paramDevicePowerInitially = pyo.Param(self.setDevice, mutable=True, within=pyo.Reals, initialize=0)
        # needed for energy storage:
        self.paramDeviceEnergyInitially = pyo.Param(self.setDevice, mutable=True, within=pyo.Reals, initialize=0)
        # target energy level at end of horizon (useful for long-term storage)
        self.paramDeviceEnergyTarget = pyo.Param(self.setDevice, mutable=True, within=pyo.Reals, initialize=0)

        # Specify handy immutable parameters (for easy access when building/updating model)
        self.paramTimestepDeltaMinutes = pyo.Param(
            within=pyo.Reals, default=self.optimisation_parameters.time_delta_minutes
        )
        self.paramTimeStorageReserveMinutes = pyo.Param(default=self.optimisation_parameters.time_reserve_minutes)
        self.paramPiecewiseRepn = pyo.Param(within=pyo.Any, default=self.optimisation_parameters.piecewise_repn)
        self.paramMaxPressureDeviation = pyo.Param(
            within=pyo.Reals,
            default=self.optimisation_parameters.max_pressure_deviation,
        )

    def _specify_variables(self):
        """specify pyomo model variables"""
        self.varEdgeFlow = pyo.Var(self.setEdge, self.setHorizon, within=pyo.Reals)
        self.varEdgeFlow12 = pyo.Var(self.setEdge, self.setHorizon, within=pyo.NonNegativeReals)
        self.varEdgeFlow21 = pyo.Var(self.setEdge, self.setHorizon, within=pyo.NonNegativeReals)
        self.varEdgeLoss = pyo.Var(self.setEdge, self.setHorizon, within=pyo.NonNegativeReals, initialize=0)
        self.varEdgeLoss12 = pyo.Var(self.setEdge, self.setHorizon, within=pyo.NonNegativeReals, initialize=0)
        self.varEdgeLoss21 = pyo.Var(self.setEdge, self.setHorizon, within=pyo.NonNegativeReals, initialize=0)
        self.varDeviceIsPrep = pyo.Var(self.setDevice, self.setHorizon, within=pyo.Binary, initialize=0)
        self.varDeviceIsOn = pyo.Var(self.setDevice, self.setHorizon, within=pyo.Binary, initialize=1)
        self.varDeviceStarting = pyo.Var(self.setDevice, self.setHorizon, within=pyo.Binary, initialize=None)
        self.varDeviceStopping = pyo.Var(self.setDevice, self.setHorizon, within=pyo.Binary, initialize=None)
        self.varDeviceStorageEnergy = pyo.Var(self.setDevice, self.setHorizon, within=pyo.Reals)
        # available reserve power from storage (linked to power rating and storage level):
        self.varDeviceStoragePmax = pyo.Var(self.setDevice, self.setHorizon, within=pyo.NonNegativeReals, initialize=0)
        # binary variable related to available powr from storage:
        self.varStorY = pyo.Var(self.setDevice, self.setHorizon, within=pyo.Binary)
        # absolute value variable for storage with target level:
        self.varDeviceStorageDeviationFromTarget = pyo.Var(self.setDevice, within=pyo.NonNegativeReals, initialize=0)
        self.varPressure = pyo.Var(
            self.setNode,
            self.setCarrier,
            self.setTerminal,
            self.setHorizon,
            within=pyo.NonNegativeReals,
            initialize=0,
        )
        self.varElVoltageAngle = pyo.Var(self.setNode, self.setHorizon, within=pyo.Reals)
        self.varDeviceFlow = pyo.Var(
            self.setDevice,
            self.setCarrier,
            self.setTerminal,
            self.setHorizon,
            within=pyo.NonNegativeReals,
            initialize=0,
        )
        self.varTerminalFlow = pyo.Var(self.setNode, self.setCarrier, self.setHorizon, within=pyo.Reals)
        # this penalty variable should only require (device,time), but the
        # piecewise constraint requires the domain to be the same as for varDeviceFlow
        self.varDevicePenalty = pyo.Var(
            self.setDevice,
            self.setCarrier,
            self.setTerminal,
            self.setHorizon,
            within=pyo.Reals,
        )

    def _specify_objective(self):
        """specify pyomo model objective"""
        obj = self.optimisation_parameters.objective
        if obj == "penalty":
            rule = self._rule_objective_penalty
        elif obj == "co2":
            rule = self._rule_objective_co2
        elif obj == "costs":
            rule = self._rule_objective_costs
        elif obj == "exportRevenue":
            rule = self._rule_objective_exportRevenue
        elif obj == "co2intensity":
            rule = self._rule_objective_co2intensity
        else:
            raise Exception("Objective '{}' has not been implemented".format(obj))
        logger.info("Using objective function: {}".format(obj))
        # logger.info(rule)
        self.objObjective = pyo.Objective(rule=rule, sense=pyo.minimize)

    def _specify_constraints(self):

        # 1. Constraints associated with each device:
        for dev in self.all_devices.values():
            list_to_reconstruct = dev.define_constraints(self)

            # Because of logic that needs to be re-evalued, these constraints need
            # to be reconstructed each optimisation:
            for constr in list_to_reconstruct:
                self.constraints_to_reconstruct.append(constr)

        # 2. Constraints associated with each node:
        for node in self.all_nodes.values():
            node.define_constraints(self)

        # 3. Constraints associated with each network type (and its edges):
        for netw in self.all_networks.values():
            netw.define_constraints(self)

        # 4. Global constraints:
        # 4.1 max limit emission rate:
        params_generic = self.optimisation_parameters
        if (params_generic.emission_rate_max is not None) and (params_generic.emission_rate_max >= 0):
            self.constr_O_emissionrate = pyo.Constraint(self.setHorizon, rule=self._rule_emissionRateLimit)
        else:
            logger.debug("No emission_rate_max limit specified")
        # 4.2 max limit emission intensity
        if (params_generic.emission_intensity_max is not None) and (params_generic.emission_intensity_max >= 0):
            self.constr_O_emissionintensity = pyo.Constraint(self.setHorizon, rule=self._rule_emissionIntensityLimit)
        else:
            logger.debug("No emission_intensity_max limit specified")
        # 4.3 electrical reserve margin:
        el_parameters = self.all_networks["el"].carrier_data
        el_reserve_margin = el_parameters.el_reserve_margin
        el_backup_margin = el_parameters.el_backup_margin
        if (el_reserve_margin is not None) and (el_reserve_margin >= 0):
            self.constr_O_elReserveMargin = pyo.Constraint(self.setHorizon, rule=self._rule_el_reserve_margin)
        else:
            logger.info("No el_reserve_margin limit specified")
        # 4.4 electrical backup power margin
        if (el_backup_margin is not None) and (el_backup_margin >= 0):
            self.constr_O_elBackupMargin = pyo.Constraint(
                self.setDevice,
                self.setHorizon,
                rule=self._rule_el_backup_margin,
            )
        else:
            logger.debug("No el_backup_margin limit specified")

    def updateOptimisationModel(self, timestep, profiles, first=False):
        """Update Pyomo model instance

        Parameters
        ----------
        timestep : int
            Present timestep
        profiles : pandas dataframe
            Time series profiles
        first : booblean
            True if it is the first time the model is updated (simulation start)
        """
        opt_timesteps = self.optimisation_parameters.optimisation_timesteps
        horizon = self.optimisation_parameters.planning_horizon
        timesteps_use_nowcast = self.optimisation_parameters.forecast_timesteps

        # Update profile (using nowcast for first 4 timesteps, forecast for rest)
        # -this is because for the first timesteps, we tend to have updated
        #  and quite good forecasts - for example, it may become apparent
        #  that there will be much less wind power than previously forecasted
        for prof in self.setProfile:
            for t in range(timesteps_use_nowcast):  # 0,1,2,3
                profile_str = "nowcast"
                if prof not in profiles["nowcast"]:
                    # no nowcast, use forecast instead
                    profile_str = "forecast"
                self.paramProfiles[prof, t] = profiles[profile_str].loc[timestep + t, prof]
            for t in range(timesteps_use_nowcast, horizon):
                self.paramProfiles[prof, t] = profiles["forecast"].loc[timestep + t, prof]

        def _updateOnTimesteps(t_prev, dev):
            # sum up consequtive timesteps starting at tprev going
            # backwards, where device has been in preparation phase
            sum_on = 0
            docontinue = True
            for tt in range(t_prev, -1, -1):
                # if (self.instance.varDeviceIsOn[dev,tt]==1):
                if pyo.value(self.varDeviceIsPrep[dev, tt]) == 1:
                    sum_on = sum_on + 1
                else:
                    docontinue = False
                    break  # exit for loop
            if docontinue:
                # we got all the way back to 0, so must include initial value
                sum_on = sum_on + self.paramDevicePrepTimestepsInitially[dev]
            return sum_on

        # Update startup/shutdown info
        # pick the last value from previous optimistion prior to the present time
        if not first:
            t_prev = opt_timesteps - 1
            for dev, dev_obj in self.all_devices.items():
                # On/off status: (round because solver doesn't alwasy return an integer)
                self.paramDeviceIsOnInitially[dev] = round(pyo.value(self.varDeviceIsOn[dev, t_prev]))
                self.paramDevicePrepTimestepsInitially[dev] = _updateOnTimesteps(t_prev, dev)
                # Initial power output (relevant for ramp rate constraint):
                if dev_obj.dev_data.max_ramp_up is not None:
                    self.paramDevicePowerInitially[dev] = dev_obj.get_flow_var(self, t_prev)
                # Energy storage:
                if dev_obj in self.devices_with_storage:
                    self.paramDeviceEnergyInitially[dev] = self.varDeviceStorageEnergy[dev, t_prev]
                    # Update target profile if present:
                    if hasattr(dev_obj.dev_data, "target_profile") and (dev_obj.dev_data.target_profile is not None):
                        prof = dev_obj.dev_data.target_profile
                        max_E = dev_obj.dev_data.E_max
                        self.paramDeviceEnergyTarget[dev] = max_E * profiles["forecast"].loc[timestep + horizon, prof]

        # These constraints need to be reconstructed to update properly
        # (pyo.value(...) and/or if sentences...)
        for c in self.constraints_to_reconstruct:
            # logger.debug("reconstructing {}".format(c))
            # c.reconstruct() <- removed in Pyomo v.6
            c.clear()
            c._constructed = False
            c.construct()
        return

    #        def storPmaxPushup(model):
    #            '''term in objective function to push varDeviceStoragePmax up
    #            to its maximum value (to get correct calculation of reserve)'''
    #            sumStorPmax=0
    #            for dev in self.setDevice:
    #                if self.paramDevice[dev]['model'] == 'storage_el':
    #                    for t in self.setHorizon:
    #                        sumStorPmax += self.varDeviceStoragePmax[dev,t]
    #            return sumStorPmax

    def _rule_objective_penalty(self, model: pyo.Model) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """'penalty' as specified through penalty functions"""
        sum_penalty = 0
        timesteps = self.setHorizon
        for d in self.setDevice:
            dev = self.all_devices[d]
            this_penalty = dev.compute_penalty(model, timesteps)
            sum_penalty = sum_penalty + this_penalty
        return sum_penalty

    def _rule_objective_co2(self, model: pyo.Model) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """CO2 emissions per sec"""
        sumE = self.compute_CO2(model)  # *self.paramParameters['CO2_price']
        return sumE

    def _rule_objective_co2intensity(self, model: pyo.Model) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """CO2 emission intensity (CO2 per exported oil/gas)
        DOES NOT WORK - NONLINEAR (ratio)"""
        sumE = self.compute_CO2_intensity(model)
        return sumE

    def _rule_objective_costs(self, model: pyo.Model) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """costs (co2 price, operating costs, startstop, storage depletaion)
        per second (assuming fixed oil/gas production)"""
        startupCosts = self.compute_startup_penalty(model)  # kr/s
        storageDepletionCosts = self.compute_costForDepletedStorage(model)
        opCosts = self.compute_operatingCosts(model)  # kr/s
        co2 = self.compute_CO2(model)  # kgCO2/s
        co2_tax = self.optimisation_parameters.co2_tax  # kr/kgCO2
        co2Cost = co2 * co2_tax  # kr/s
        sumCost = co2Cost + startupCosts + storageDepletionCosts + opCosts
        return sumCost

    def _rule_objective_exportRevenue(self, model: pyo.Model) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """revenue from exported oil and gas minus costs (co2 price and
        operating costs) per second"""
        sumRevenue = self.compute_exportRevenue(model)  # kr/s
        startupCosts = self.compute_startup_penalty(model)  # kr/s
        co2 = self.compute_CO2(model)  # kgCO2/s
        co2_tax = self.optimisation_parameters.co2_tax  # kr/kgCO2
        co2Cost = co2 * co2_tax  # kr/s
        storageDepletionCosts = self.compute_costForDepletedStorage(model)
        opCosts = self.compute_operatingCosts(model)  # kr/s
        sumCost = -sumRevenue + co2Cost + startupCosts + storageDepletionCosts + opCosts
        return sumCost

    def _rule_emissionRateLimit(self, model: pyo.Model, t) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """Upper limit on CO2 emission rate"""
        params_generic = self.optimisation_parameters
        emissionRateMax = params_generic.emission_rate_max
        lhs = self.compute_CO2(model, timesteps=[t])
        rhs = emissionRateMax
        return pyo.Expression(lhs <= rhs)

    def _rule_emissionIntensityLimit(self, model: pyo.Model, t) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """Upper limit on CO2 emission intensity"""
        params_generic = self.optimisation_parameters
        emissionIntensityMax = params_generic.emission_intensity_max
        lhs = self.compute_CO2(model, timesteps=[t])
        rhs = emissionIntensityMax * self.compute_oilgas_export(model, timesteps=[t])
        return pyo.Expression(lhs <= rhs)

    def _rule_el_reserve_margin(self, pyomo_model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """Reserve margin constraint (electrical supply)
        Not used capacity by power suppliers/storage/load flexibility
        must be larger than some specified margin
        (to cope with unforeseen variations)
        """
        # exclude constraint for first timesteps since the point of the
        # dispatch margin is exactly to cope with forecast errors
        if t < self.optimisation_parameters.forecast_timesteps:
            return pyo.Constraint.Skip

        network_el = self.all_networks["el"]

        # Fixme: Error in type hint.
        margin = network_el.carrier_data.el_reserve_margin
        capacity_unused = network_el.compute_el_reserve(pyomo_model=pyomo_model, t=t, all_devices=self.all_devices)
        expr = capacity_unused >= margin
        return expr

    def _rule_el_backup_margin(self, model: pyo.Model, dev, t) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """Backup capacity constraint (electrical supply)
        Not used capacity by other online power suppliers plus sheddable
        load must be larger than power output of this device
        (to take over in case of a fault)

        elBackupMargin is zero or positive (if loss of load is acceptable)
        """
        dev_obj = self.all_devices[dev]
        network_el = self.all_networks["el"]
        margin = network_el.carrier_data.el_backup_margin
        if "el" not in dev_obj.carrier_out:
            # this is not a power generator
            return pyo.Constraint.Skip
        res_otherdevs = network_el.compute_el_reserve(model, t, self.all_devices, exclude_device=dev)
        expr = res_otherdevs - self.varDeviceFlow[dev, "el", "out", t] >= -margin
        return expr

    def compute_CO2(self, model: pyo.Model, devices=None, timesteps=None):
        """compute CO2 emissions - average per sec (kgCO2/s)"""
        if devices is None:
            devices = self.setDevice
        if timesteps is None:
            timesteps = self.setHorizon
        sumCO2 = 0
        for d in devices:
            dev = self.all_devices[d]
            thisCO2 = dev.compute_CO2(model, timesteps)
            sumCO2 = sumCO2 + thisCO2

        # Average per s
        return sumCO2 / len(timesteps)

    def compute_CO2_intensity(self, model: pyo.Model, timesteps=None):
        """CO2 emission per exported oil/gas (kgCO2/Sm3oe)"""
        if timesteps is None:
            timesteps = self.setHorizon

        co2_kg_per_time = self.compute_CO2(model, devices=None, timesteps=timesteps)
        flow_oil_equivalents_m3_per_time = self.compute_oilgas_export(model, timesteps)
        if pyo.value(flow_oil_equivalents_m3_per_time) != 0:
            return co2_kg_per_time / flow_oil_equivalents_m3_per_time
        elif pyo.value(flow_oil_equivalents_m3_per_time) == 0:
            # logger.debug("zero export, so co2 intensity set to None")
            return None
        else:
            return co2_kg_per_time

    def compute_startup_penalty(self, model: pyo.Model, devices=None, timesteps=None):
        """startup costs (average per sec)"""
        if timesteps is None:
            timesteps = self.setHorizon
        if devices is None:
            devices = self.setDevice
        start_stop_costs = 0
        for d in devices:
            dev_obj = self.all_devices[d]
            thisCost = dev_obj.compute_startup_penalty(model, timesteps)
            start_stop_costs += thisCost
        # get average per sec:
        deltaT = self.optimisation_parameters.time_delta_minutes * 60
        sumTime = len(timesteps) * deltaT
        start_stop_costs = start_stop_costs / sumTime
        return start_stop_costs

    logger.info("TODO: operating cost for el storage - needs improvement")

    def compute_operatingCosts(self, model: pyo.Model):
        """term in objective function to represent fuel costs or similar
        as average per sec ($/s)

        opCost = energy costs (NOK/MJ, or NOK/Sm3)
        Note: el costs per MJ not per MWh
        """
        sumCost = 0
        timesteps = self.setHorizon
        for dev in self.setDevice:
            dev_obj = self.all_devices[dev]
            thisCost = dev_obj.compute_operating_costs(model, timesteps)
            sumCost += thisCost
        return sumCost

    def compute_costForDepletedStorage(self, model: pyo.Model):
        """term in objective function to discourage depleting battery,
        making sure it is used only when required"""
        storCost = 0
        timesteps = self.setHorizon
        for dev in self.setDevice:
            dev_obj = self.all_devices[dev]
            thisCost = dev_obj.compute_cost_for_depleted_storage(model, timesteps)
            storCost += thisCost
        return storCost

    def compute_exportRevenue(self, model: pyo.Model, carriers=None, timesteps=None):
        """revenue from exported oil and gas - average per sec ($/s)"""
        return self.compute_export(model, value="revenue", carriers=carriers, timesteps=timesteps)

    def compute_oilgas_export(self, model: pyo.Model, timesteps=None):
        """Export volume (Sm3oe/s)"""
        return self.compute_export(model, value="volume", carriers=["oil", "gas"], timesteps=timesteps)

    def compute_export(self, model: pyo.Model, value="revenue", carriers=None, timesteps=None):
        """Compute average export (volume or revenue)

        Parameters
        ----------
        model : oogeso model
        value : string ("revenue", "volume")
            which value to compute, revenue (€/s) or volume (Sm3oe/s)

        Computes the energy/mass flow into (sink) devices with a price.CARRIER
        parameter defined (CARRIER can be any of 'oil', 'gas', 'el')
        """
        if carriers is None:
            carriers = self.setCarrier
        if timesteps is None:
            timesteps = self.setHorizon

        sumValue = 0
        for dev in self.setDevice:
            dev_obj = self.all_devices[dev]
            thisValue = dev_obj.compute_export(model, value, carriers, timesteps)
            sumValue += thisValue
        # average per second (timedelta is not required)
        sumValue = sumValue / len(timesteps)
        return sumValue

    def getDevicesInout(self, carrier_in=None, carrier_out=None):
        """devices that have the specified connections in and out"""
        devs = []
        for d, dev_obj in self.all_devices.items():
            ok_in = (carrier_in is None) or (carrier_in in dev_obj.carrier_in)
            ok_out = (carrier_out is None) or (carrier_out in dev_obj.carrier_out)
            if ok_in and ok_out:
                devs.append(d)
        return devs

    def extract_all_variable_values(self, timelimit: int = None, timeshift: int = 0) -> Dict[str, pd.Series]:
        """Extract variable values and return as a dictionary of pandas milti-index series"""
        all_vars = [
            self.varEdgeFlow,
            self.varEdgeLoss,
            self.varDeviceIsPrep,
            self.varDeviceIsOn,
            self.varDeviceStarting,
            self.varDeviceStopping,
            self.varDeviceStorageEnergy,
            self.varDeviceStoragePmax,
            self.varPressure,
            self.varElVoltageAngle,
            self.varDeviceFlow,
            self.varTerminalFlow,
            self.varDevicePenalty,
        ]
        # all_vars = self.component_objects(pyo.Var, active=True)
        all_values = {}
        for myvar in all_vars:
            # extract the variable index names in the right order
            indices = [index_set.doc for index_set in myvar._implicit_subsets]
            var_values = myvar.get_values()
            if not var_values:
                # print("var_values=", var_values)
                # empty dictionary, so no variables to store
                all_values[myvar.name] = None
                continue
            # This creates a pandas.Series:
            df = pd.DataFrame.from_dict(var_values, orient="index", columns=["value"])["value"]
            df.index = pd.MultiIndex.from_tuples(df.index, names=indices)
            # check that all vales are non-negative for deviceflow and give warning otherwise
            if (myvar == self.varDeviceFlow) and (df < -self.ZERO_WARNING_THRESHOLD).any():
                ind = df[df < -self.ZERO_WARNING_THRESHOLD].index[0]
                logger.warning("Negative number in varDeviceFlow - set to zero ({}:{})".format(ind, df[ind]))
                df = df.clip(lower=0)

            # ignore NA values
            df = df.dropna()
            if df.empty:
                all_values[myvar.name] = None
                continue

            if timelimit is not None:
                mask = df.index.get_level_values("time") < timelimit
                df = df[mask]
            if timeshift > 0:
                level_time = df.index.names.index("time")
                new_values = df.index.levels[level_time] + timeshift
                new_index = df.index.set_levels(new_values, level="time")
                df.index = new_index

            all_values[myvar.name] = df
        return all_values



from .plots import (
    plot_CO2_intensity,
    plot_CO2_rate,
    plot_CO2_rate_per_dev,
    plot_device_power_energy,
    plot_device_power_flow_pressure,
    plot_device_profile,
    plot_df,
    plot_el_backup,
    plot_el_backup2,
    plot_export_revenue,
    plot_gas_turbine_efficiency,
    plot_network,
    plot_profiles,
    plot_reserve,
    plot_sum_power_mix,
)
